@@ -4,25 +4,21 @@
  *	Apply Function to Files
  */
 
-#include <signal.h>
 #include <fcntl.h>
 #include "fd.h"
 #include "term.h"
 #include "func.h"
 #include "kanji.h"
 
-#if	!MSDOS
-#include <sys/file.h>
-#include <sys/param.h>
-#endif
+#define	MAXTIMESTR	8
 
-typedef struct _attr_t {
-	u_short mode;
+typedef struct _attrib_t {
+	u_int mode;
 #ifdef	HAVEFLAGS
 	u_long flags;
 #endif
-	char timestr[2][9];
-} attr_t;
+	char timestr[2][MAXTIMESTR + 1];
+} attrib_t;
 
 #ifndef	_NODOSDRIVE
 extern int preparedrv __P_((int));
@@ -68,16 +64,19 @@ static int touchdir __P_((char *));
 #ifndef	_NOEXTRACOPY
 static int mvdir __P_((char *));
 #endif
-static VOID NEAR showattr __P_((namelist *, attr_t *, int y));
+static VOID NEAR showattr __P_((namelist *, attrib_t *, int y));
 static char **NEAR getdirtree __P_((char *, char **, int *, int));
 static int NEAR _applydir __P_((char *, int (*)(char *),
-		int (*)(char *), int (*)(char *), int, char *));
+		int (*)(char *), int (*)(char *), int, char *, int));
+static int forcecpfile __P_((char *));
+static int forcecpdir __P_((char *));
+static int forcetouchdir __P_((char *));
 
 int copypolicy = 0;
 int removepolicy = 0;
 char *destpath = NULL;
 
-static u_short attrmode = 0;
+static u_int attrmode = 0;
 #ifdef	HAVEFLAGS
 static u_long attrflags = 0;
 #endif
@@ -129,11 +128,17 @@ char *path, *org;
 static int NEAR islowerdir(path, org)
 char *path, *org;
 {
+#if	!MSDOS && !defined (_NODOSDRIVE)
+	char orgpath[MAXPATHLEN];
+#endif
 	char *cp, *top, *cwd, tmp[MAXPATHLEN];
 	int i;
 
 	cwd = getwd2();
 	if (!org) org = cwd;
+#if	!MSDOS && !defined (_NODOSDRIVE)
+	else org = nodospath(orgpath, org);
+#endif
 	strcpy(tmp, path);
 	top = tmp;
 #if	MSDOS || !defined (_NODOSDRIVE)
@@ -271,7 +276,7 @@ struct stat *stp1, *stp2;
 
 	for (;;) {
 		if (!n || n == 2) {
-			locate(0, LCMDLINE);
+			locate(0, L_CMDLINE);
 			putterm(l_clear);
 			putch2('[');
 			cp = SAMEF_K;
@@ -333,7 +338,7 @@ struct stat *stp1, *stp2;
 			case 2:
 				if ((cp = strrdelim(dest, 1))) cp++;
 				else cp = dest;
-				lcmdline = LINFO - n_line;
+				lcmdline = L_INFO;
 				if (!(tmp = inputstr(NEWNM_K, 1,
 					-1, NULL, -1))) return(-1);
 				strcpy(cp, tmp);
@@ -357,6 +362,7 @@ struct stat *stp1, *stp2;
 			case 5:
 				if (stp1 -> st_mtime < stp2 -> st_mtime)
 					return(-1);
+/*FALLTHRU*/
 			case 6:
 				strcatdelim2(path, forwardpath, file);
 				cp = strrdelim(path, 0);
@@ -381,6 +387,9 @@ struct stat *stp1, *stp2;
 				break;
 		}
 	}
+
+/*NOTREACHED*/
+	return(0);
 }
 
 static int NEAR checkrmv(path, mode)
@@ -394,6 +403,7 @@ int mode;
 	if (Xaccess(path, mode) >= 0) return(0);
 #else	/* !MSDOS */
 	struct stat *stp, st;
+	char *tmp, dir[MAXPATHLEN];
 
 # ifndef	_NODOSDRIVE
 	if (dospath2(path)) {
@@ -403,9 +413,7 @@ int mode;
 	else
 # endif
 	{
-		char *tmp, dir[MAXPATHLEN];
-
-		if (_Xlstat(path, &st) < 0) {
+		if (nodoslstat(path, &st) < 0) {
 			warning(-1, path);
 			return(-1);
 		}
@@ -415,12 +423,12 @@ int mode;
 		else if (tmp == path) strcpy(dir, _SS_);
 		else strncpy2(dir, path, tmp - path);
 
-		if (_Xlstat(dir, &st) < 0) {
+		if (nodoslstat(dir, &st) < 0) {
 			warning(-1, dir);
 			return(-1);
 		}
 		stp = &st;
-		if (access(path, mode) >= 0) return(0);
+		if (Xaccess(path, mode) >= 0) return(0);
 	}
 #endif	/* !MSDOS */
 	if (errno == ENOENT) return(0);
@@ -443,7 +451,7 @@ int mode;
 	}
 #endif
 	if (removepolicy > 0) return(removepolicy - 2);
-	locate(0, LCMDLINE);
+	locate(0, L_CMDLINE);
 	putterm(l_clear);
 	putch2('[');
 	cp = DELPM_K;
@@ -465,8 +473,8 @@ int mode;
 static int safecopy(path)
 char *path;
 {
-	char dest[MAXPATHLEN];
 	struct stat st1, st2;
+	char dest[MAXPATHLEN];
 
 	if (checkdupl(path, dest, &st1, &st2) < 0) return(0);
 	if (safecpfile(path, dest, &st1, &st2) < 0) return(-1);
@@ -476,8 +484,8 @@ char *path;
 static int safemove(path)
 char *path;
 {
-	char dest[MAXPATHLEN];
 	struct stat st1, st2;
+	char dest[MAXPATHLEN];
 
 	if (checkdupl(path, dest, &st1, &st2) < 0) return(0);
 	if (checkrmv(path, W_OK) < 0) return(1);
@@ -488,19 +496,20 @@ char *path;
 static int cpdir(path)
 char *path;
 {
-	char dest[MAXPATHLEN];
 	struct stat st1, st2;
+	char dest[MAXPATHLEN];
 
 	switch (checkdupl(path, dest, &st1, &st2)) {
 		case 2:
 		/* Already exist, but not directory */
 			if (Xunlink(dest) < 0) return(-1);
+/*FALLTHRU*/
 		case 0:
 		/* Not exist */
 #if	MSDOS
-			if (Xmkdir(dest, st1.st_mode | S_IWRITE) < 0)
+			if (Xmkdir(dest, (st1.st_mode & 0777) | S_IWRITE) < 0)
 #else
-			if (Xmkdir(dest, st1.st_mode) < 0)
+			if (Xmkdir(dest, st1.st_mode & 0777) < 0)
 #endif
 				return(-1);
 			destmtime = st1.st_mtime;
@@ -518,7 +527,7 @@ char *path;
 			break;
 	}
 	if (destdir) free(destdir);
-	destdir = strdup2(dest + strlen(destpath) + 1);
+	destdir = strdup2(&(dest[strlen(destpath) + 1]));
 	return(0);
 }
 
@@ -527,8 +536,8 @@ static int touchdir(path)
 char *path;
 {
 #if	MSDOS || !defined (_NOEXTRACOPY)
-	char dest[MAXPATHLEN];
 	struct stat st;
+	char dest[MAXPATHLEN];
 
 # if	!MSDOS
 	if (!inheritcopy) return(0);
@@ -538,6 +547,9 @@ char *path;
 		st.st_mtime = destmtime;
 		st.st_atime = destatime;
 	}
+# ifdef	HAVEFLAGS
+	st.st_flags = 0xffffffff;
+# endif
 	if (touchfile(dest, &st) < 0) return(-1);
 #endif	/* MSDOS || !_NOEXTRACOPY */
 	return(0);
@@ -547,14 +559,17 @@ char *path;
 static int mvdir(path)
 char *path;
 {
-	char dest[MAXPATHLEN];
 	struct stat st;
+	char dest[MAXPATHLEN];
 
 	if (getdestpath(path, dest, &st) < 0) return(-2);
 	if (destmtime != (time_t)-1 || destatime != (time_t)-1) {
 		st.st_mtime = destmtime;
 		st.st_atime = destatime;
 	}
+# ifdef	HAVEFLAGS
+	st.st_flags = 0xffffffff;
+# endif
 	if (touchfile(dest, &st) < 0) return(-1);
 	if (checkrmv(path, R_OK | W_OK | X_OK) < 0) return(1);
 	return(Xrmdir(path));
@@ -585,7 +600,7 @@ char *path;
 
 	if (regexp_exec(findregexp, cp, 1)) {
 		if (path[0] == '.' && path[1] == _SC_) path += 2;
-		locate(0, LCMDLINE);
+		locate(0, L_CMDLINE);
 		putterm(l_clear);
 		putch2('[');
 		kanjiputs2(path, n_column - 2, -1);
@@ -617,70 +632,81 @@ char *path;
 
 static VOID NEAR showattr(listp, attr, y)
 namelist *listp;
-attr_t *attr;
+attrib_t *attr;
 int y;
 {
 	struct tm *tm;
 	char buf[WMODE + 1];
+	int x1, x2, w;
 
 	tm = localtime(&(listp -> st_mtim));
+	if (isbestomit()) {
+		x1 = n_column / 2 - 16;
+		x2 = n_column / 2 - 3;
+		w = 12;
+	}
+	else {
+		x1 = n_column / 2 - 20;
+		x2 = n_column / 2;
+		w = 16;
+	}
 
 	locate(0, y);
 	putterm(l_clear);
 
 	locate(0, ++y);
 	putterm(l_clear);
-	locate(n_column / 2 - 20, y);
+	locate(x1, y);
 	putch2('[');
-	kanjiputs2(listp -> name, 16, 0);
+	kanjiputs2(listp -> name, w, 0);
 	putch2(']');
-	locate(n_column / 2 + 3, y);
+	locate(x2 + 3, y);
 	kanjiputs(TOLD_K);
-	locate(n_column / 2 + 13, y);
+	locate(x2 + 13, y);
 	kanjiputs(TNEW_K);
 
 	locate(0, ++y);
 	putterm(l_clear);
-	locate(n_column / 2 - 20, y);
+	locate(x1, y);
 	kanjiputs(TMODE_K);
-	locate(n_column / 2, y);
+	locate(x2, y);
 	putmode(buf, listp -> st_mode);
-	cputs2(buf + 1);
-	locate(n_column / 2 + 10, y);
+	cputs2(&(buf[1]));
+	locate(x2 + 10, y);
 	putmode(buf, attr -> mode);
-	cputs2(buf + 1);
+	cputs2(&(buf[1]));
 
 #ifdef	HAVEFLAGS
 	locate(0, ++y);
 	putterm(l_clear);
-	locate(n_column / 2 - 20, y);
+	locate(x1, y);
 	kanjiputs(TFLAG_K);
-	locate(n_column / 2, y);
+	locate(x2, y);
 	putflags(buf, listp -> st_flags);
 	cputs2(buf);
-	locate(n_column / 2 + 10, y);
+	locate(x2 + 10, y);
 	putflags(buf, attr -> flags);
 	cputs2(buf);
 #endif
 
 	locate(0, ++y);
 	putterm(l_clear);
-	locate(n_column / 2 - 20, y);
+	locate(x1, y);
 	kanjiputs(TDATE_K);
-	locate(n_column / 2, y);
+	locate(x2, y);
 	cprintf2("%02d-%02d-%02d",
 		tm -> tm_year % 100, tm -> tm_mon + 1, tm -> tm_mday);
-	locate(n_column / 2 + 10, y);
+	locate(x2 + 10, y);
 	cputs2(attr -> timestr[0]);
 
 	locate(0, ++y);
 	putterm(l_clear);
-	locate(n_column / 2 - 20, y);
+	locate(x1, y);
 	kanjiputs(TTIME_K);
-	locate(n_column / 2, y);
+	locate(x2, y);
 	cprintf2("%02d:%02d:%02d",
 		tm -> tm_hour, tm -> tm_min, tm -> tm_sec);
-	locate(n_column / 2 + 10, y);
+	locate(x2 + 10, y);
 	cputs2(attr -> timestr[1]);
 
 	locate(0, ++y);
@@ -689,34 +715,32 @@ int y;
 
 int inputattr(listp, flag)
 namelist *listp;
-u_short flag;
+int flag;
 {
 	struct tm *tm;
-	attr_t attr;
+	attrib_t attr;
 	char buf[WMODE + 1];
-	u_short tmp;
-	int i, ch, x, y, yy, ymin, ymax, dupwin_x, dupwin_y;
+	u_int tmp;
+	int i, ch, x, y, xx, yy, ymin, ymax, dupwin_x, dupwin_y;
 
 	dupwin_x = win_x;
 	dupwin_y = win_y;
 	subwindow = 1;
-#ifndef	_NOEDITMODE
-	Xgetkey(-1);
-#endif
+	Xgetkey(-1, 0);
 
 	yy = LFILETOP;
-#ifndef	_NOSPLITWIN
-	while (n_line - yy < WMODELINE + 6) yy--;
-#endif
+	while (yy + WMODELINE + 5 > n_line - 1) yy--;
+	if (yy <= L_TITLE) yy = L_TITLE + 1;
+	xx = n_column / 2 + ((isbestomit()) ? 7 : 10);
 
 	attr.mode = listp -> st_mode;
 #ifdef	HAVEFLAGS
 	attr.flags = listp -> st_flags;
 #endif
 	tm = localtime(&(listp -> st_mtim));
-	sprintf(attr.timestr[0], "%02d-%02d-%02d",
+	snprintf2(attr.timestr[0], sizeof(attr.timestr[0]), "%02d-%02d-%02d",
 		tm -> tm_year % 100, tm -> tm_mon + 1, tm -> tm_mday);
-	sprintf(attr.timestr[1], "%02d:%02d:%02d",
+	snprintf2(attr.timestr[1], sizeof(attr.timestr[1]), "%02d:%02d:%02d",
 		tm -> tm_hour, tm -> tm_min, tm -> tm_sec);
 	showattr(listp, &attr, yy);
 	y = ymin = (flag & 1) ? 0 : WMODELINE;
@@ -724,13 +748,13 @@ u_short flag;
 	x = 0;
 
 	do {
-		win_x = n_column / 2 + 10 + x;
+		win_x = xx + x;
 		win_y = yy + y + 2;
 		locate(win_x, win_y);
 		tflush();
 
 		keyflush();
-		switch (ch = Xgetkey(SIGALRM)) {
+		switch (ch = Xgetkey(1, 0)) {
 			case K_UP:
 				if (y > ymin) y--;
 				else y = ymax;
@@ -774,6 +798,7 @@ u_short flag;
 				if (y < WMODELINE) break;
 				putch2(ch);
 				attr.timestr[y - WMODELINE][x] = ch;
+/*FALLTHRU*/
 			case K_RIGHT:
 #ifdef	HAVEFLAGS
 				if (y == WMODELINE - 1) {
@@ -814,14 +839,18 @@ u_short flag;
 				}
 				else if (x > 0) x--;
 				break;
-			case CTRL('L'):
+			case K_CTRL('L'):
+				yy = LFILETOP;
+				while (yy + WMODELINE + 5 > n_line - 1) yy--;
+				if (yy <= L_TITLE) yy = L_TITLE + 1;
+				xx = n_column / 2 + ((isbestomit()) ? 7 : 10);
 				showattr(listp, &attr, yy);
 				break;
 			case ' ':
 #ifdef	HAVEFLAGS
 				if (y == WMODELINE - 1) {
 					attr.flags ^= flaglist[x];
-					locate(n_column / 2 + 10, yy + y + 2);
+					locate(xx, yy + y + 2);
 					putflags(buf, attr.flags);
 					cputs2(buf);
 					break;
@@ -847,9 +876,9 @@ u_short flag;
 				}
 				attr.mode ^= tmp;
 #endif
-				locate(n_column / 2 + 10, yy + y + 2);
+				locate(xx, yy + y + 2);
 				putmode(buf, attr.mode);
-				cputs2(buf + 1);
+				cputs2(&(buf[1]));
 				break;
 			default:
 				break;
@@ -859,9 +888,7 @@ u_short flag;
 	win_x = dupwin_x;
 	win_y = dupwin_y;
 	subwindow = 0;
-#ifndef	_NOEDITMODE
-	Xgetkey(-1);
-#endif
+	Xgetkey(-1, 0);
 
 	if (ch == K_ESC) return(0);
 
@@ -920,10 +947,21 @@ int applyfile(func, endmes)
 int (*func)__P_((char *));
 char *endmes;
 {
+#if	!MSDOS && !defined (_NODOSDRIVE)
+	char path[MAXPATHLEN];
+#endif
 	int i, ret, old, dupfilepos;
 
 	dupfilepos = filepos;
 	ret = old = filepos;
+
+	if (mark <= 0) {
+		i = (*func)(fnodospath(path, filepos));
+		if (i < 0) warning(-1, filelist[filepos].name);
+		else if (!i) ret++;
+		return(ret);
+	}
+
 	helpbar();
 	for (filepos = 0; filepos < maxfile; filepos++) {
 		if (intrkey()) {
@@ -935,9 +973,9 @@ char *endmes;
 		movepos(old, 0);
 		locate(win_x, win_y);
 		tflush();
-		if ((i = (*func)(filelist[filepos].name)) < 0)
-			warning(-1, filelist[filepos].name);
-		if (!i && ret == filepos) ret++;
+		i = (*func)(fnodospath(path, filepos));
+		if (i < 0) warning(-1, filelist[filepos].name);
+		else if (!i && ret == filepos) ret++;
 
 		old = filepos;
 	}
@@ -958,39 +996,36 @@ int *maxp, depth;
 	struct dirent *dp;
 	struct stat st;
 	char *cp;
+	int d;
 
-	if (!(dirp = Xopendir(dir))) {
-		warning(-1, dir);
-		if (dirlist) free(dirlist);
-		return(NULL);
-	}
+	if (!(dirp = Xopendir(dir))) return(dirlist);
 	cp = strcatdelim(dir);
 	while ((dp = Xreaddir(dirp))) {
 		if (isdotdir(dp -> d_name)) continue;
 		strcpy(cp, dp -> d_name);
 
-		if (Xlstat(dir, &st) >= 0
-		&& (st.st_mode & S_IFMT) == S_IFDIR) {
-			dirlist = b_realloc(dirlist, *maxp, char *);
-			dirlist[(*maxp)++] = strdup2(dir);
-			if (!depth || depth > 1) {
-				dirlist = getdirtree(dir,
-					dirlist, maxp, --depth);
-				if (!dirlist) return(NULL);
-			}
-		}
+		if (Xlstat(dir, &st) < 0 || (st.st_mode & S_IFMT) != S_IFDIR)
+			continue;
+		dirlist = b_realloc(dirlist, *maxp, char *);
+		dirlist[(*maxp)++] = strdup2(dir);
+		if (!(d = depth)) /*EMPTY*/;
+		else if (d > 1) d--;
+		else continue;
+		dirlist = getdirtree(dir, dirlist, maxp, d);
 	}
+	if (cp > dir) *(cp - 1) = '\0';
 	Xclosedir(dirp);
 	return(dirlist);
 }
 
-static int NEAR _applydir(dir, funcf, funcd1, funcd2, order, endmes)
+static int NEAR _applydir(dir, funcf, funcd1, funcd2, order, endmes, verbose)
 char *dir;
 int (*funcf)__P_((char *));
 int (*funcd1)__P_((char *));
 int (*funcd2)__P_((char *));
 int order;
 char *endmes;
+int verbose;
 {
 	DIR *dirp;
 	struct dirent *dp;
@@ -1001,19 +1036,26 @@ char *endmes;
 
 	if (intrkey()) return(-2);
 
-	locate(0, LCMDLINE);
-	putterm(l_clear);
-	putch2('[');
-	cp = (dir[0] == '.' && dir[1] == _SC_) ? dir + 2 : dir;
-	kanjiputs2(cp, n_column - 2, -1);
-	putch2(']');
-	tflush();
+	if (verbose) {
+		locate(0, L_CMDLINE);
+		putterm(l_clear);
+		putch2('[');
+		cp = (dir[0] == '.' && dir[1] == _SC_) ? dir + 2 : dir;
+		kanjiputs2(cp, n_column - 2, -1);
+		putch2(']');
+		tflush();
+	}
+#ifdef	FAKEUNINIT
+	else cp = NULL;	/* fake for -Wuninitialized */
+#endif
 
 	if (!funcd1 && !funcd2) order = 0;
-	fname = strcatdelim2(path, dir, NULL);
+	strcpy(path, dir);
 
 	ret = max = 0;
-	dirlist = getdirtree(path, NULL, &max, (order == 2) ? 0 : 1);
+	if (!order) dirlist = NULL;
+	else dirlist = getdirtree(path, NULL, &max, (order == 2) ? 0 : 1);
+	fname = strcatdelim(path);
 
 	destmtime = destatime = (time_t)-1;
 	if ((order == 1 || order == 2) && (ret = (*funcd1)(dir)) < 0) {
@@ -1029,45 +1071,51 @@ char *endmes;
 
 	if (order) for (ndir = 0; ndir < max; ndir++) {
 		if (order != 2) ret = _applydir(dirlist[ndir], funcf,
-				funcd1, funcd2, order, NULL);
+				funcd1, funcd2, order, NULL, verbose);
 		else if ((ret = (*funcd1)(dirlist[ndir])) < 0) {
 			if (ret == -1) warning(-1, dir);
 			ret = -2;
 		}
 		else ret = _applydir(dirlist[ndir], funcf,
-			funcd1, funcd2, 0, NULL);
+			funcd1, funcd2, 0, NULL, verbose);
 		if (ret < -1) {
 			for (; ndir < max; ndir++) free(dirlist[ndir]);
 			free(dirlist);
 			return(ret);
 		}
 		free(dirlist[ndir]);
-		locate(0, LCMDLINE);
-		putterm(l_clear);
-		putch2('[');
-		kanjiputs2(cp, n_column - 2, -1);
-		putch2(']');
-		tflush();
+
+		if (verbose) {
+			locate(0, L_CMDLINE);
+			putterm(l_clear);
+			putch2('[');
+			kanjiputs2(cp, n_column - 2, -1);
+			putch2(']');
+			tflush();
+		}
 	}
 	if (dirlist) free(dirlist);
 
-	if (!(dirp = Xopendir(dir))) {
-		warning(-1, dir);
-		return(-1);
-	}
-	while ((dp = Xreaddir(dirp))) {
-		if (intrkey()) return(-2);
-		if (isdotdir(dp -> d_name)) continue;
-		strcpy(fname, dp -> d_name);
+	if (!(dirp = Xopendir(dir))) warning(-1, dir);
+	else {
+		while ((dp = Xreaddir(dirp))) {
+			if (intrkey()) {
+				ret = -2;
+				break;
+			}
+			if (isdotdir(dp -> d_name)) continue;
+			strcpy(fname, dp -> d_name);
 
-		if (Xlstat(path, &st) < 0) warning(-1, path);
-		else if ((st.st_mode & S_IFMT) == S_IFDIR) continue;
-		else if ((ret = (*funcf)(path)) < 0) {
-			if (ret == -1) warning(-1, path);
-			else return(ret);
+			if (Xlstat(path, &st) < 0) warning(-1, path);
+			else if ((st.st_mode & S_IFMT) == S_IFDIR) continue;
+			else if ((ret = (*funcf)(path)) < 0) {
+				if (ret < -1) break;
+				warning(-1, path);
+			}
 		}
+		Xclosedir(dirp);
+		if (ret < -1) return(ret);
 	}
-	Xclosedir(dirp);
 
 	destmtime = dupmtime;
 	destatime = dupatime;
@@ -1085,12 +1133,21 @@ int order;
 char *endmes;
 {
 	char path[MAXPATHLEN];
+	int verbose;
 
-	if (dir[0] == '.' && dir[1] == '.' && !dir[2]) {
+	verbose = 1;
+	if (!dir) {
+		dir = ".";
+		verbose = 0;
+	}
+	else if (dir[0] == '.' && dir[1] == '.' && !dir[2]) {
 		realpath2(dir, path, 0);
 		dir = path;
 	}
-	return(_applydir(dir, funcf, funcd1, funcd2, order, endmes));
+#if	!MSDOS && !defined (_NODOSDRIVE)
+	else dir = nodospath(path, dir);
+#endif
+	return(_applydir(dir, funcf, funcd1, funcd2, order, endmes, verbose));
 }
 
 /*ARGSUSED*/
@@ -1118,9 +1175,10 @@ int tr;
 #ifndef	_NODOSDRIVE
 	if (dospath3("")) waitmes();
 #endif
-	if (mark > 0) applyfile(safecopy, ENDCP_K);
-	else if (isdir(&(filelist[filepos]))
-	&& !islink(&(filelist[filepos]))) {
+	if (mark > 0
+	|| (!isdir(&(filelist[filepos])) || islink(&(filelist[filepos]))))
+		applyfile(safecopy, ENDCP_K);
+	else {
 #ifdef	_NOEXTRACOPY
 		if (copypolicy) {
 			warning(EEXIST, filelist[filepos].name);
@@ -1132,8 +1190,7 @@ int tr;
 			(islowerdir(destpath, filelist[filepos].name)) ? 2 : 1,
 			ENDCP_K);
 	}
-	else if (safecopy(filelist[filepos].name) < 0)
-		warning(-1, filelist[filepos].name);
+
 	free(destpath);
 	if (destdir) free(destdir);
 #ifndef	_NOEXTRACOPY
@@ -1177,21 +1234,30 @@ int tr;
 	if (mark > 0) filepos = applyfile(safemove, ENDMV_K);
 	else if (islowerdir(destpath, filelist[filepos].name))
 		warning(EINVAL, filelist[filepos].name);
-	else if ((i = safemove(filelist[filepos].name)) < 0) {
-#ifdef	_NOEXTRACOPY
-		warning(-1, filelist[filepos].name);
-#else
-		if (errno != EXDEV
-		|| !isdir(&(filelist[filepos]))
-		|| islink(&(filelist[filepos])))
-			warning(-1, filelist[filepos].name);
-		else if (applydir(filelist[filepos].name,
-			safemove, cpdir, mvdir,
-			(islowerdir(destpath, filelist[filepos].name)) ? 2 : 1,
-			ENDMV_K) >= 0) filepos++;
+	else {
+#if	!MSDOS && !defined (_NODOSDRIVE)
+		char path[MAXPATHLEN];
 #endif
+
+		i = safemove(fnodospath(path, filepos));
+		if (!i) filepos++;
+		else if (i < 0) {
+#ifdef	_NOEXTRACOPY
+			warning(-1, filelist[filepos].name);
+#else
+			if (errno != EXDEV
+			|| !isdir(&(filelist[filepos]))
+			|| islink(&(filelist[filepos])))
+				warning(-1, filelist[filepos].name);
+			else if (applydir(filelist[filepos].name,
+				safemove, cpdir, mvdir,
+				(islowerdir(destpath, filelist[filepos].name))
+				? 2 : 1,
+				ENDMV_K) >= 0) filepos++;
+#endif
+		}
 	}
-	else if (!i) filepos++;
+
 	if (filepos >= maxfile) filepos = maxfile - 1;
 	free(destpath);
 	if (destdir) free(destdir);
@@ -1208,3 +1274,64 @@ int tr;
 #endif
 	return(4);
 }
+
+static int forcecpfile(path)
+char *path;
+{
+	struct stat st;
+	char dest[MAXPATHLEN];
+
+	if (getdestpath(path, dest, &st) < 0) return(0);
+	if (safecpfile(path, dest, &st, NULL) < 0) return(-1);
+#ifdef	HAVEFLAGS
+	st.st_flags = 0xffffffff;
+#endif
+	return(touchfile(dest, &st));
+}
+
+static int forcecpdir(path)
+char *path;
+{
+	struct stat st;
+	char dest[MAXPATHLEN];
+
+	if (isdotdir(path)) return(0);
+	if (getdestpath(path, dest, &st) < 0) return(0);
+#if	MSDOS
+	while (Xmkdir(dest, (st.st_mode & 0777) | S_IWRITE) < 0)
+#else
+	while (Xmkdir(dest, st.st_mode & 0777) < 0)
+#endif
+	{
+		if (errno != EEXIST) return(-1);
+		if (Xstat(dest, &st) < 0) return(-1);
+		if ((st.st_mode & S_IFMT) == S_IFDIR) break;
+		if (Xunlink(dest) < 0) return(-1);
+	}
+	return(0);
+}
+
+static int forcetouchdir(path)
+char *path;
+{
+	struct stat st;
+	char dest[MAXPATHLEN];
+
+	if (isdotdir(path)) return(0);
+	if (getdestpath(path, dest, &st) < 0) return(-2);
+#ifdef	HAVEFLAGS
+	st.st_flags = 0xffffffff;
+#endif
+	return(touchfile(dest, &st));
+}
+
+int forcemovefile(dest)
+char *dest;
+{
+	int ret;
+
+	destpath = dest;
+	ret = applydir(NULL, forcecpfile, forcecpdir, forcetouchdir, 1, NULL);
+	destpath = NULL;
+	return(ret);
+}
