@@ -18,14 +18,13 @@
 #include <process.h>
 extern int getcurdrv __P_((VOID_A));
 # ifndef	_NOUSELFN
-extern char *preparefile __P_((char *, char *, int));
+extern char *preparefile __P_((char *, char *));
 # endif
 #else
 #include <pwd.h>
 #include <grp.h>
-#include <sys/file.h>
-#include <sys/param.h>
 #endif
+
 #ifndef	_NODOSDRIVE
 extern int flushdrv __P_((int, VOID_T (*)__P_((VOID_A))));
 extern int shutdrv __P_((int));
@@ -41,29 +40,56 @@ extern off_t totalsize;
 extern off_t blocksize;
 extern char fullpath[];
 extern char *tmpfilename;
-
-#define	DOSDIRENT		32
-#define	LFNENTSIZ		13
-#ifndef	O_BINARY
-#define	O_BINARY		0
+extern int physical_path;
+extern int noconv;
+#ifndef	_NODOSDRIVE
+extern int dosdrive;
 #endif
 
-#define	realdirsiz(name)	((dos) ? ((k_strlen(name) / LFNENTSIZ + 1) \
-					* DOSDIRENT + DOSDIRENT) \
-				:	(((strlen(name) + boundary) \
-					& ~(boundary - 1)) + dirsize))
-#define	getnamlen(ent)		((dos) ? ((((ent) - DOSDIRENT) \
-					/ DOSDIRENT) * LFNENTSIZ - 1) \
-				: 	((((ent) - dirsize) \
-					& ~(boundary - 1)) - 1))
+#define	MAXTMPNAMLEN	8
+#ifndef	O_BINARY
+#define	O_BINARY	0
+#endif
+#ifdef	_NODOSDRIVE
+#define	DOSDIRENT	32
+#define	LFNENTSIZ	13
+#endif
+#define	LFNONLY		" +,;=[]"
+#define	DOSBODYLEN	8
+#define	DOSEXTLEN	3
 
+#ifndef	L_SET
+# ifdef	SEEK_SET
+# define	L_SET	SEEK_SET
+# else
+# define	L_SET	0
+# endif
+#endif
+
+#ifdef	_NODOSDRIVE
+#define	nodoschdir	Xchdir
+#define	nodosgetwd	Xgetwd
+#define	nodosmkdir	Xmkdir
+#define	nodosrmdir	Xrmdir
+#else
+static int NEAR nodoschdir __P_((char *));
+static char *NEAR nodosgetwd __P_((char *));
+static int NEAR nodosmkdir __P_((char *, int));
+static int NEAR nodosrmdir __P_((char *));
+#endif
 static int NEAR cpfile __P_((char *, char *, struct stat *, struct stat *));
 static VOID changemes __P_((VOID_A));
+static int NEAR genrand __P_((int));
+static int dounlink __P_((char *));
 #ifndef	_NOWRITEFS
-static int NEAR k_strlen __P_((char *));
+static int NEAR isexist __P_((char *));
+static int NEAR realdirsiz __P_((char *, int, int, int, int));
+static int NEAR getnamlen __P_((int, int, int, int, int));
+static int NEAR saferename __P_((char *, char *));
 static char *NEAR maketmpfile __P_((int, int, char *, char *));
 #if	!MSDOS
-static char *NEAR getentnum __P_((char *, int, int));
+static off_t NEAR getdirblocksize __P_((char *));
+static u_char *NEAR getentnum __P_((char *, off_t));
 #endif
 static VOID NEAR restorefile __P_((char *, char *, int));
 #endif	/* !_NOWRITEFS */
@@ -72,18 +98,31 @@ char *deftmpdir = NULL;
 #if	!MSDOS && !defined (_NOEXTRACOPY)
 int inheritcopy = 0;
 #endif
+int tmpumask = TMPUMASK;
 
 #ifndef	_NODOSDRIVE
 static int dosdrv = -1;
 #endif
 
 
+#if	!MSDOS && !defined (_NODOSDRIVE)
+char *nodospath(path, file)
+char *path, *file;
+{
+	if (!_dospath(file)) return(file);
+	path[0] = '.';
+	path[1] = _SC_;
+	strcpy(&(path[2]), file);
+	return(path);
+}
+#endif
+
 #if	MSDOS
 int logical_access(mode)
-u_short mode;
+u_int mode;
 #else
 int logical_access(mode, uid, gid)
-u_short mode;
+u_int mode;
 uid_t uid;
 gid_t gid;
 #endif
@@ -160,22 +199,14 @@ gid_t gid;
 int getstatus(list)
 namelist *list;
 {
+#if	!MSDOS && !defined (_NODOSDRIVE)
+	char path[MAXPATHLEN];
+#endif
 	struct stat st, lst;
+	char *cp;
 
-#if	MSDOS
-	if (stat2(list -> name, &st) < 0) return(-1);
-	memcpy((char *)&lst, (char *)&st, sizeof(struct stat));
-#else	/* !MSDOS */
-# ifndef	_NODOSDRIVE
-	if (dospath(list -> name, NULL)) {
-		if (stat2(list -> name, &st) < 0) return(-1);
-		memcpy((char *)&lst, (char *)&st, sizeof(struct stat));
-	}
-	else
-# endif
-	if (Xlstat(list -> name, &lst) < 0
-	|| stat2(list -> name, &st) < 0) return(-1);
-#endif	/* !MSDOS */
+	cp = nodospath(path, list -> name);
+	if (Xlstat(cp, &lst) < 0 || stat2(cp, &st) < 0) return(-1);
 	list -> flags = 0;
 	if ((st.st_mode & S_IFMT) == S_IFDIR) list -> flags |= F_ISDIR;
 	if ((lst.st_mode & S_IFMT) == S_IFLNK) list -> flags |= F_ISLNK;
@@ -194,7 +225,7 @@ namelist *list;
 #else
 	list -> st_uid = lst.st_uid;
 	list -> st_gid = lst.st_gid;
-	list -> st_size = isdev(list) ? lst.st_rdev : lst.st_size;
+	list -> st_size = isdev(list) ? (off_t)(lst.st_rdev) : lst.st_size;
 #endif
 #ifdef	HAVEFLAGS
 	list -> st_flags = lst.st_flags;
@@ -202,9 +233,9 @@ namelist *list;
 	list -> st_mtim = lst.st_mtime;
 	list -> flags |=
 #if	MSDOS
-		logical_access((u_short)(st.st_mode));
+		logical_access((u_int)(st.st_mode));
 #else
-		logical_access((u_short)(st.st_mode), st.st_uid, st.st_gid);
+		logical_access((u_int)(st.st_mode), st.st_uid, st.st_gid);
 #endif
 
 	if (isfile(list) && list -> st_size) {
@@ -215,21 +246,13 @@ namelist *list;
 	return(0);
 }
 
-int isdotdir(name)
-char *name;
-{
-	if (name[0] == '.'
-	&& (!name[1] || (name[1] == '.' && !name[2]))) return(1);
-	return(0);
-}
-
 int cmplist(vp1, vp2)
 CONST VOID_P vp1;
 CONST VOID_P vp2;
 {
 	namelist *list1, *list2;
 	char *cp1, *cp2;
-	long tmp;
+	int tmp;
 
 	list1 = (namelist *)vp1;
 	list2 = (namelist *)vp2;
@@ -254,7 +277,8 @@ CONST VOID_P vp2;
 		case 5:
 			tmp = (int)strlen(list1 -> name)
 				- (int)strlen(list2 -> name);
-			if (tmp != 0) break;
+			if (tmp) break;
+/*FALLTHRU*/
 		case 1:
 			tmp = strpathcmp2(list1 -> name, list2 -> name);
 			break;
@@ -287,11 +311,16 @@ CONST VOID_P vp2;
 			if (isdir(list1))
 				tmp = strpathcmp2(list1 -> name,
 					list2 -> name);
-			else tmp = (long)(list1 -> st_size)
-				- (long)(list2 -> st_size);
+			else if (list1 -> st_size == list2 -> st_size)
+				tmp = 0;
+			else tmp = (list1 -> st_size > list2 -> st_size)
+				? 1 : -1;
 			break;
 		case 4:
-			tmp = list1 -> st_mtim - list2 -> st_mtim;
+			if (list1 -> st_mtim == list2 -> st_mtim)
+				tmp = 0;
+			else tmp = (list1 -> st_mtim > list2 -> st_mtim)
+				? 1 : -1;
 			break;
 		default:
 			tmp = 0;
@@ -301,9 +330,7 @@ CONST VOID_P vp2;
 	if (sorton > 7) tmp = -tmp;
 	if (!tmp) tmp = list1 -> ent - list2 -> ent;
 
-	if (tmp > 0) return(1);
-	if (tmp < 0) return(-1);
-	return(0);
+	return(tmp);
 }
 
 #ifndef	_NOTREE
@@ -397,9 +424,10 @@ char *buf;
 #endif
 		_chdir2(cp) < 0) return(-1);
 		homedir = getwd2();
-		if (_chdir2(cwd) < 0) error(cwd);
+		if (_chdir2(cwd) < 0) lostcwd(cwd);
 	}
 
+	if (buf && !physical_path) strcpy(cwd, fullpath);
 	len = strlen(homedir);
 #if	MSDOS
 	if (len <= 3 || strnpathcmp(cwd, homedir, len)) {
@@ -425,32 +453,36 @@ char *dir;
 {
 	struct stat st;
 	char tmp[MAXPATHLEN];
-#if	MSDOS
 	char *cp;
-#endif
 
-	if (stat2(dir, &st) < 0) {
-		if (errno != ENOENT) return(-1);
-		if (Xmkdir(dir, 0777) < 0) {
-#if	MSDOS
-			if (errno != EEXIST) return(-1);
-			cp = (_dospath(dir)) ? dir + 2 : dir;
-			if (cp[0] != _SC_ || cp[1]) {
-				if ((cp = strrdelim(dir, 1))) cp++;
-				else cp = dir;
-				if (cp[0] != '.' || cp[1]) return(-1);
-			}
-			st.st_mode = S_IFDIR;
-#else
-			return(-1);
+	cp = dir;
+#if	MSDOS || !defined (_NODOSDRIVE)
+	if (_dospath(dir)) cp += 2;
 #endif
+	if (!isdotdir(cp)) {
+		if (stat2(dir, &st) < 0) {
+			if (errno != ENOENT) return(-1);
+			if (Xmkdir(dir, 0777) < 0) {
+#if	MSDOS
+				if (errno != EEXIST) return(-1);
+				if (cp[0] != _SC_ || cp[1]) {
+					if ((cp = strrdelim(dir, 1))) cp++;
+					else cp = dir;
+					if (cp[0] != '.' || cp[1]) return(-1);
+				}
+				st.st_mode = S_IFDIR;
+#else
+				return(-1);
+#endif
+			}
+			else if (stat2(dir, &st) < 0) return(-1);
+		}
+		if ((st.st_mode & S_IFMT) != S_IFDIR) {
+			errno = ENOTDIR;
+			return(-1);
 		}
-		else if (stat2(dir, &st) < 0) return(-1);
-	}
-	if ((st.st_mode & S_IFMT) != S_IFDIR) {
-		errno = ENOTDIR;
-		return(-1);
 	}
+
 	entryhist(1, realpath2(dir, tmp, 0), 1);
 	return(0);
 }
@@ -466,6 +498,9 @@ struct stat *stp;
 	if ((st.st_mode & S_IFMT) == S_IFLNK) return(1);
 
 	ret = 0;
+#ifdef	FAKEUNINIT
+	duperrno = errno;	/* fake for -Wuninitialized */
+#endif
 #if	MSDOS
 	if (!(st.st_mode & S_IWRITE)) Xchmod(path, (st.st_mode | S_IWRITE));
 #endif
@@ -492,11 +527,11 @@ struct stat *stp;
 		}
 #endif
 	}
-#ifdef HAVEFLAGS
+#ifdef	HAVEFLAGS
 	if (stp -> st_flags != 0xffffffff) {
 # ifndef	_NODOSDRIVE
 		if (dospath(path, NULL)) {
-			errno = EACCES;
+			duperrno = EACCES;
 			ret = -1;
 		}
 		else
@@ -520,14 +555,83 @@ struct stat *stp;
 	else if (!(st.st_mode & S_IWRITE)) Xchmod(path, st.st_mode);
 #else
 	if (stp -> st_gid != (gid_t)-1) {
+# ifndef	_NODOSDRIVE
+		if (dospath(path, NULL)) /*EMPTY*/;
+		else
+# endif
 		if (chown(path, stp -> st_uid, stp -> st_gid) < 0)
-			chown(path, -1, stp -> st_gid);
+			chown(path, (uid_t)-1, stp -> st_gid);
 	}
 #endif
 	if (ret < 0) errno = duperrno;
 	return(ret);
 }
 
+#ifndef	_NODOSDRIVE
+static int NEAR nodoschdir(path)
+char *path;
+{
+	int n, dupdosdrive;
+
+	dupdosdrive = dosdrive;
+	dosdrive = 0;
+	n = Xchdir(path);
+	dosdrive = dupdosdrive;
+	return(n);
+}
+
+static char *NEAR nodosgetwd(path)
+char *path;
+{
+	char *cp;
+	int dupdosdrive;
+
+	dupdosdrive = dosdrive;
+	dosdrive = 0;
+	cp = Xgetwd(path);
+	dosdrive = dupdosdrive;
+	return(cp);
+}
+
+int nodoslstat(path, stp)
+char *path;
+struct stat *stp;
+{
+	int n, dupdosdrive;
+
+	dupdosdrive = dosdrive;
+	dosdrive = 0;
+	n = Xlstat(path, stp);
+	dosdrive = dupdosdrive;
+	return(n);
+}
+
+static int NEAR nodosmkdir(path, mode)
+char *path;
+int mode;
+{
+	int n, dupdosdrive;
+
+	dupdosdrive = dosdrive;
+	dosdrive = 0;
+	n = Xmkdir(path, mode);
+	dosdrive = dupdosdrive;
+	return(n);
+}
+
+static int NEAR nodosrmdir(path)
+char *path;
+{
+	int n, dupdosdrive;
+
+	dupdosdrive = dosdrive;
+	dosdrive = 0;
+	n = Xrmdir(path);
+	dosdrive = dupdosdrive;
+	return(n);
+}
+#endif	/* !_NODOSDRIVE */
+
 VOID lostcwd(path)
 char *path;
 {
@@ -537,16 +641,14 @@ char *path;
 	duperrno = errno;
 	if (!path) path = buf;
 
-	if (path != fullpath && !rawchdir(fullpath) && _Xgetwd(path))
+	if (path != fullpath && !nodoschdir(fullpath) && nodosgetwd(path))
 		cp = NOCWD_K;
-	else {
-		if ((cp = gethomedir()) && !rawchdir(cp) && _Xgetwd(path))
-			cp = GOHOM_K;
-		else if (!rawchdir(_SS_) && _Xgetwd(path)) cp = GOROT_K;
-		else error(_SS_);
-		strncpy2(fullpath, path, MAXPATHLEN - 1);
-	}
+	else if ((cp = gethomedir()) && !nodoschdir(cp) && nodosgetwd(path))
+		cp = GOHOM_K;
+	else if (!nodoschdir(_SS_) && nodosgetwd(path)) cp = GOROT_K;
+	else error(_SS_);
 
+	if (path != fullpath) strncpy2(fullpath, path, MAXPATHLEN - 1);
 	warning(0, cp);
 	errno = duperrno;
 }
@@ -560,12 +662,12 @@ int size;
 
 	n = Xwrite(fd, buf, size);
 #if	MSDOS
-	if (n < size) n = 0;
-#else
-	if (n < 0) {
-		if (errno == EINTR) n = size;
-		else if (errno == ENOSPC) n = 0;
+	if (n >= 0 && n < size) {
+		n = -1;
+		errno = ENOSPC;
 	}
+#else
+	if (n < 0 && errno == EINTR) n = size;
 #endif
 	return(n);
 }
@@ -579,7 +681,7 @@ struct stat *stp1, *stp2;
 	struct stat st;
 #endif
 	char buf[BUFSIZ];
-	int i, n, fd1, fd2, tmperrno;
+	int i, fd1, fd2, duperrno;
 
 #if	MSDOS
 	if (!stp2 && Xlstat(dest, &st) >= 0) stp2 = &st;
@@ -591,41 +693,43 @@ struct stat *stp1, *stp2;
 		buf[i] = '\0';
 		return(Xsymlink(buf, dest));
 	}
-	if ((fd1 = Xopen(src, O_RDONLY | O_BINARY, stp1 -> st_mode)) < 0)
+	if ((fd1 = Xopen(src, O_BINARY | O_RDONLY, stp1 -> st_mode)) < 0)
 		return(-1);
-	if ((fd2 = Xopen(dest, O_WRONLY | O_BINARY | O_CREAT | O_TRUNC,
 #if	MSDOS
-	stp1 -> st_mode | S_IWRITE)) < 0) {
+	fd2 = Xopen(dest,
+		O_BINARY | O_WRONLY | O_CREAT | O_TRUNC,
+		stp1 -> st_mode | S_IWRITE);
 #else
-	stp1 -> st_mode)) < 0) {
+	fd2 = Xopen(dest,
+		O_BINARY | O_WRONLY | O_CREAT | O_TRUNC, stp1 -> st_mode);
 #endif
+	if (fd2 < 0) {
 		Xclose(fd1);
 		return(-1);
 	}
 
 	for (;;) {
 		while ((i = Xread(fd1, buf, BUFSIZ)) < 0 && errno == EINTR);
-		tmperrno = errno;
+		duperrno = errno;
 		if (i < BUFSIZ) break;
-		if ((n = safewrite(fd2, buf, BUFSIZ)) <= 0) {
-			tmperrno = (n < 0) ? errno : ENOSPC;
-			i = -1;
+		if ((i = safewrite(fd2, buf, BUFSIZ)) < 0) {
+			duperrno = errno;
 			break;
 		}
 	}
-	if (i > 0 && (n = safewrite(fd2, buf, i)) <= 0) {
-		tmperrno = (n < 0) ? errno : ENOSPC;
-		i = -1;
-	}
+	if (i > 0 && (i = safewrite(fd2, buf, i)) < 0) duperrno = errno;
 
 	Xclose(fd2);
 	Xclose(fd1);
 
 	if (i < 0) {
 		Xunlink(dest);
-		errno = tmperrno;
+		errno = duperrno;
 		return(-1);
 	}
+#ifdef	HAVEFLAGS
+	stp1 -> st_flags = 0xffffffff;
+#endif
 #if	MSDOS
 	if (touchfile(dest, stp1) < 0) return(-1);
 #else
@@ -650,7 +754,7 @@ struct stat *stp1, *stp2;
 #endif
 
 	for (;;) {
-		if (cpfile(src, dest, stp1, stp2) >= 0) return(0);
+		if (cpfile(src, dest, stp1, stp2) >= 0) break;
 		if (errno && errno != ENOSPC) return(-1);
 		for (;;) {
 			if (!yesno(NOSPC_K)) {
@@ -667,6 +771,8 @@ struct stat *stp1, *stp2;
 			break;
 		}
 	}
+
+	return(0);
 }
 
 int safemvfile(src, dest, stp1, stp2)
@@ -676,51 +782,146 @@ struct stat *stp1, *stp2;
 	if (Xrename(src, dest) >= 0) return(0);
 	if (errno != EXDEV || (stp1 -> st_mode & S_IFMT) == S_IFDIR)
 		return(-1);
-	if (safecpfile(src, dest, stp1, stp2) < 0 || Xunlink(src) < 0
-	|| touchfile(dest, stp1) < 0) return(-1);
-	return(0);
+	if (safecpfile(src, dest, stp1, stp2) < 0 || Xunlink(src) < 0)
+		return(-1);
+#ifdef	HAVEFLAGS
+	stp1 -> st_flags = 0xffffffff;
+#endif
+	return (touchfile(dest, stp1));
+}
+
+static int NEAR genrand(max)
+int max;
+{
+	static long last = -1L;
+	time_t now;
+
+	if (last < 0L) {
+		now = time2();
+		last = ((now & 0xff) << 16) + (now & ~0xff) + getpid();
+	}
+
+	do {
+		last = last * (u_long)1103515245 + 12345;
+	} while (last < 0L);
+
+	return((last / 65537L) % max);
+}
+
+char *genrandname(buf, len)
+char *buf;
+int len;
+{
+	static char seq[] = {
+#if	MSDOS
+		'0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
+		'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J',
+		'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',
+		'U', 'V', 'W', 'X', 'Y', 'Z', '_'
+#else
+		'0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
+		'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',
+		'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
+		'u', 'v', 'w', 'x', 'y', 'z', '_'
+#endif
+	};
+	int i, j, c;
+
+	if (!buf) {
+		for (i = 0; i < sizeof(seq) / sizeof(char); i++) {
+			j = genrand(sizeof(seq) / sizeof(char));
+			c = seq[i];
+			seq[i] = seq[j];
+			seq[j] = c;
+		}
+	}
+	else {
+		for (i = 0; i < len; i++) {
+			j = genrand(sizeof(seq) / sizeof(char));
+			buf[i] = seq[j];
+		}
+		buf[i] = '\0';
+	}
+	return(buf);
 }
 
 int mktmpdir(dir)
 char *dir;
 {
 	char *cp, path[MAXPATHLEN];
-	int no;
+	int n, len, mask;
 
-	if (!deftmpdir || !*deftmpdir || !tmpfilename || !*tmpfilename
-	|| !dir || !*dir) {
+	if (!deftmpdir || !*deftmpdir || !dir) {
 		errno = ENOENT;
 		return(-1);
 	}
 	realpath2(deftmpdir, path, 1);
 	free(deftmpdir);
 #if	MSDOS && !defined (_NODOSDRIVE)
-	if (checkdrive(toupper2(path[0]) - 'A'))
-		if (!_Xgetwd(path)) {
-			lostcwd(path);
-			deftmpdir = NULL;
-			return(-1);
-		}
+	if (checkdrive(toupper2(path[0]) - 'A') && !nodosgetwd(path)) {
+		lostcwd(path);
+		deftmpdir = NULL;
+		return(-1);
+	}
 #endif
 	deftmpdir = strdup2(path);
 #if	MSDOS
-	no = getcurdrv();
-	*path = (no >= 'A' && no <= 'Z') ? toupper2(*path) : tolower2(*path);
+	n = getcurdrv();
+	*path = (n >= 'A' && n <= 'Z') ? toupper2(*path) : tolower2(*path);
 #endif
-	strcpy(strcatdelim(path), tmpfilename);
-	if (_Xmkdir(path, 0777) < 0 && errno != EEXIST) return(-1);
-	strcpy((cp = strcatdelim(path)), dir);
-	if (_Xmkdir(path, 0777) < 0 && errno != EEXIST) {
-		*(--cp) = '\0';
-		no = errno;
-		if (_Xrmdir(path) < 0
-		&& errno != ENOTEMPTY && errno != EEXIST && errno != EACCES)
-			error(path);
-		errno = no;
-		return(-1);
+
+	mask = 0777 & ~tmpumask;
+	cp = strcatdelim(path);
+	if (tmpfilename) {
+		strcpy(cp, tmpfilename);
+		if (Xaccess(path, R_OK | W_OK | X_OK) < 0) {
+			free(tmpfilename);
+			tmpfilename = NULL;
+		}
 	}
-	strcpy(dir, path);
-	return(0);
+	if (!tmpfilename) {
+		n = sizeof(TMPPREFIX) - 1;
+		strncpy(cp, TMPPREFIX, n);
+		len = sizeof(path) - 1 - (cp - path);
+		if (len > MAXTMPNAMLEN) len = MAXTMPNAMLEN;
+		len -= n;
+		genrandname(NULL, 0);
+
+		for (;;) {
+			genrandname(cp + n, len);
+			if (nodosmkdir(path, mask) >= 0) break;
+			if (errno != EEXIST) return(-1);
+		}
+		tmpfilename = strdup2(cp);
+	}
+
+	if (!(n = strlen(dir))) {
+		strcpy(dir, path);
+		return(0);
+	}
+
+	strncpy((cp = strcatdelim(path)), dir, n);
+	len = sizeof(path) - 1 - (cp - path);
+	if (len > MAXTMPNAMLEN) len = MAXTMPNAMLEN;
+	len -= n;
+	genrandname(NULL, 0);
+
+	for (;;) {
+		genrandname(cp + n, len);
+		if (nodosmkdir(path, mask) >= 0) {
+			strcpy(dir, path);
+			return(0);
+		}
+		if (errno != EEXIST) break;
+	}
+
+	if (cp > path) {
+		*(--cp) = '\0';
+		n = errno;
+		nodosrmdir(path);
+		errno = n;
+	}
+	return(-1);
 }
 
 int rmtmpdir(dir)
@@ -728,75 +929,103 @@ char *dir;
 {
 	char path[MAXPATHLEN];
 
-	if (dir && *dir && _Xrmdir(dir) < 0) return(-1);
+	if (dir && *dir && nodosrmdir(dir) < 0) return(-1);
 	if (!deftmpdir || !*deftmpdir || !tmpfilename || !*tmpfilename) {
 		errno = ENOENT;
 		return(-1);
 	}
 	strcatdelim2(path, deftmpdir, tmpfilename);
-	if (_Xrmdir(path) < 0
-	&& errno != ENOTEMPTY && errno != EEXIST && errno != EACCES)
+	if (nodosrmdir(path) >= 0) {
+		free(tmpfilename);
+		tmpfilename = NULL;
+	}
+	else if (errno != ENOTEMPTY && errno != EEXIST && errno != EACCES)
 		return(-1);
 	return(0);
 }
 
-VOID removetmp(dir, subdir, file)
-char *dir, *subdir, *file;
+int mktmpfile(file)
+char *file;
 {
-	char *cp, *tmp, *dupdir;
+	char *cp, path[MAXPATHLEN];
+	int fd, len, duperrno;
 
-	if (!dir || !*dir) subdir = file = NULL;
+	path[0] = '\0';
+	if (mktmpdir(path) < 0) return(-1);
+	cp = strcatdelim(path);
+	len = sizeof(path) - 1 - (cp - path);
+	if (len > MAXTMPNAMLEN) len = MAXTMPNAMLEN;
+	genrandname(NULL, 0);
+
+	for (;;) {
+		genrandname(cp, len);
+		fd = Xopen(path, O_BINARY | O_WRONLY | O_CREAT | O_EXCL,
+			0666 & ~tmpumask);
+		if (fd >= 0) {
+			strcpy(file, path);
+			return(fd);
+		}
+		if (errno != EEXIST) break;
+	}
+	duperrno = errno;
+	rmtmpdir(NULL);
+	errno = duperrno;
+	return(-1);
+}
+
+int rmtmpfile(file)
+char *file;
+{
+	if ((Xunlink(file) < 0 && errno != ENOENT) || rmtmpdir(NULL) < 0)
+		return(-1);
+	return(0);
+}
+
+static int dounlink(path)
+char *path;
+{
+	return(Xunlink(path));
+}
+
+static int dormdir(path)
+char *path;
+{
+	if (isdotdir(path)) return(0);
+	return(Xrmdir(path));
+}
+
+VOID removetmp(dir, file)
+char *dir, *file;
+{
+#if	!MSDOS && !defined (_NODOSDRIVE)
+	char path[MAXPATHLEN];
+#endif
+
+	if (!dir || !*dir) /*EMPTY*/;
 	else if (_chdir2(dir) < 0) {
 		warning(-1, dir);
 		*dir = '\0';
-		subdir = file = NULL;
 	}
-	if (subdir) {
-		while (*subdir == _SC_) subdir++;
-		if (*subdir && _chdir2(subdir) < 0) {
-			warning(-1, subdir);
-			subdir = file = NULL;
-		}
-	}
-	if (file) {
-		if (*file) {
-			if (Xunlink(file) < 0) {
-				warning(-1, file);
-				subdir = NULL;
-			}
-		}
-		else if (applydir(".", Xunlink, NULL, NULL, 0, NULL) < 0) {
-			warning(-1, dir);
-			subdir = NULL;
-		}
-	}
-	if (subdir && *subdir) {
-		if (_chdir2(dir) < 0) error(dir);
-		dupdir = strdup2(subdir);
-		cp = dupdir + strlen(dupdir);
-		for (;;) {
-			*cp = '\0';
-			if ((tmp = strrdelim(dupdir, 1))) tmp++;
-			else tmp = dupdir;
-			if ((*tmp && strcmp(tmp, ".") && _Xrmdir(dupdir) < 0)
-			|| tmp == dupdir) break;
-			cp = tmp - 1;
-		}
-		free(dupdir);
-	}
-#if	MSDOS && !defined (_NODOSDRIVE)
-	if (_chdir2(deftmpdir) < 0) error(deftmpdir);
-#endif
+	else if (!file) applydir(NULL, dounlink, NULL, dormdir, 3, NULL);
+	else if (Xunlink(nodospath(path, file)) < 0) warning(-1, file);
+
 #ifndef	_NODOSDRIVE
 	if (dosdrv >= 0) {
-		if (lastdrv >= 0) shutdrv(lastdrv);
+		if (lastdrv >= 0) {
+# if	MSDOS
+			if (_chdir2(deftmpdir) < 0);
+			else
+# endif
+			shutdrv(lastdrv);
+		}
 		lastdrv = dosdrv;
 		dosdrv = -1;
 	}
 #endif
+
 	if (_chdir2(fullpath) < 0) lostcwd(fullpath);
 	if (dir) {
-		if (rmtmpdir(dir) < 0) warning(-1, dir);
+		if (*dir && rmtmpdir(dir) < 0) warning(-1, dir);
 		free(dir);
 	}
 }
@@ -805,7 +1034,7 @@ int forcecleandir(dir, file)
 char *dir, *file;
 {
 #if	!MSDOS
-	long pid;
+	p_id_t pid;
 #endif
 	extern char **environ;
 	char buf[MAXPATHLEN];
@@ -818,7 +1047,7 @@ char *dir, *file;
 #if	MSDOS
 	spawnlpe(P_WAIT, "DELTREE.EXE", "DELTREE", "/Y", buf, NULL, environ);
 #else
-	if ((pid = fork()) < 0) return(-1);
+	if ((pid = fork()) < 0L) return(-1);
 	else if (!pid) {
 		execle("/bin/rm", "rm", "-rf", buf, NULL, environ);
 		_exit(1);
@@ -828,24 +1057,36 @@ char *dir, *file;
 }
 
 #ifndef	_NODOSDRIVE
+char *dostmpdir(drive)
+int drive;
+{
+	char path[MAXPATHLEN];
+
+	path[0] = DOSTMPPREFIX;
+	path[1] = drive;
+	path[2] = '\0';
+	if (mktmpdir(path) < 0) return(NULL);
+	return(strdup2(path));
+}
+
 int tmpdosdupl(dir, dirp, single)
 char *dir, **dirp;
 int single;
 {
 	struct stat st;
-	char *cp, path[MAXPATHLEN];
+	char *cp, *tmpdir, path[MAXPATHLEN];
+# if	!MSDOS
+	char tmp[MAXPATHLEN];
+# endif
 	int i, drive;
 
 	if (!(drive = dospath2(dir))) return(0);
-	cp = path;
-	*(cp++) = 'D';
-	*(cp++) = drive;
-	*cp = '\0';
-	if (mktmpdir(path) < 0) {
-		warning(-1, path);
+	if (!(tmpdir = dostmpdir(drive))) {
+		warning(-1, dir);
 		return(-1);
 	}
 
+	strcpy(path, tmpdir);
 	cp = strcatdelim(path);
 	waitmes();
 
@@ -853,9 +1094,9 @@ int single;
 		strcpy(cp, filelist[filepos].name);
 		st.st_mode = filelist[filepos].st_mode;
 		st.st_atime = st.st_mtime = filelist[filepos].st_mtim;
-		if (cpfile(filelist[filepos].name, path, &st, NULL) < 0) {
-			*(--cp) = '\0';
-			removetmp(path, NULL, NULL);
+		if (cpfile(fnodospath(tmp, filepos), path, &st, NULL) < 0) {
+			warning(-1, path);
+			removetmp(tmpdir, NULL);
 			return(-1);
 		}
 	}
@@ -864,18 +1105,21 @@ int single;
 		strcpy(cp, filelist[i].name);
 		st.st_mode = filelist[i].st_mode;
 		st.st_atime = st.st_mtime = filelist[i].st_mtim;
-		if (cpfile(filelist[i].name, path, &st, NULL) < 0) {
-			*(--cp) = '\0';
-			removetmp(path, NULL, "");
+		if (cpfile(fnodospath(tmp, i), path, &st, NULL) < 0) {
+			warning(-1, path);
+			removetmp(tmpdir, NULL);
 			return(-1);
 		}
 	}
 
-	*(--cp) = '\0';
 	dosdrv = lastdrv;
 	lastdrv = -1;
-	if (_chdir2(path) < 0) error(path);
-	*dirp = strdup2(path);
+	if (_chdir2(tmpdir) < 0) {
+		warning(-1, tmpdir);
+		removetmp(tmpdir, NULL);
+		return(-1);
+	}
+	*dirp = tmpdir;
 	return(drive);
 }
 
@@ -885,10 +1129,16 @@ char *file;
 {
 	struct stat st;
 	char path[MAXPATHLEN];
+# if	!MSDOS
+	char tmp[MAXPATHLEN];
+# endif
 
 	path[0] = drive;
 	path[1] = ':';
-	strcpy(path + 2, file);
+	strcpy(&(path[2]), file);
+# if	!MSDOS
+	file = nodospath(tmp, file);
+# endif
 	waitmes();
 	if (Xlstat(file, &st) < 0 || cpfile(file, path, &st, NULL) < 0)
 		return(-1);
@@ -897,61 +1147,124 @@ char *file;
 }
 #endif	/* !_NODOSDRIVE */
 
-int isexist(file)
+#ifndef	_NOWRITEFS
+static int NEAR isexist(file)
 char *file;
 {
-#if	MSDOS && !defined (_NOUSELFN)
-	char *cp, buf[MAXPATHLEN];
+	struct stat st;
 
-	if (!(cp = preparefile(file, buf, 0)) && errno == ENOENT) return(0);
-	if (cp != file) return(1);
-#endif
-	if (Xaccess(file, F_OK) < 0 && errno == ENOENT) return(0);
+	if (Xlstat(file, &st) < 0 && errno == ENOENT) return(0);
 	return(1);
 }
 
-#ifndef	_NOWRITEFS
-static int NEAR k_strlen(s)
+static int NEAR realdirsiz(s, dos, boundary, dirsize, ofs)
 char *s;
+int dos, boundary, dirsize, ofs;
 {
-	int i;
+	int i, len, lfn, dot;
 
-	for (i = 0; s[i]; i++) if (iskanji1(s, i)) i++;
-	return(i);
+	if (!dos) {
+		len = (strlen(s) + ofs + boundary) & ~(boundary - 1);
+		return(len + dirsize);
+	}
+
+	if (dos > 1 && !isdotdir(s)) {
+
+		lfn = dot = 0;
+		for (i = len = 0; s[i]; i++, len++) {
+			if (s[i] == '.') {
+				if (dot || !i || i > DOSBODYLEN) lfn = 1;
+				dot = i + 1;
+			}
+			else if (issjis1(s[i]) && issjis2(s[i + 1])) {
+				i++;
+				lfn = 1;
+			}
+			else if (!lfn && strchr(LFNONLY, s[i])) lfn = 1;
+		}
+		if (lfn) /*EMPTY*/;
+		else if (dot) {
+			if (i - dot > DOSEXTLEN) lfn = 1;
+		}
+		else if (i > DOSBODYLEN) lfn = 1;
+
+		len += ofs;
+		if (lfn) return((len / LFNENTSIZ + 1) * DOSDIRENT + DOSDIRENT);
+	}
+	return(DOSDIRENT);
 }
 
+static int NEAR getnamlen(size, dos, boundary, dirsize, ofs)
+int size, dos, boundary, dirsize, ofs;
+{
+	if (!dos) {
+		size -= dirsize;
+		return((size & ~(boundary - 1)) - 1 - ofs);
+	}
+
+	if (size > DOSDIRENT) {
+		size -= DOSDIRENT;
+		return((size / DOSDIRENT) * LFNENTSIZ - 1);
+	}
+	return(DOSBODYLEN);
+}
+
+static int NEAR saferename(from, to)
+char *from, *to;
+{
+#if	!MSDOS && !defined (_NODOSDRIVE)
+	char fpath[MAXPATHLEN], tpath[MAXPATHLEN];
+#endif
+
+	if (!strpathcmp(from, to)) return(0);
+#if	!MSDOS && !defined (_NODOSDRIVE)
+	from = nodospath(fpath, from);
+	to = nodospath(tpath, to);
+#endif
+	return(Xrename(from, to));
+}
+
+/*ARGSUSED*/
 static char *NEAR maketmpfile(len, dos, tmpdir, old)
 int len, dos;
 char *tmpdir, *old;
 {
 	char *fname, path[MAXPATHLEN];
-	int i, l, fd;
+	int l, fd;
 
 	if (len < 0) return(NULL);
 	fname = malloc2(len + 1);
-
-	memset(fname, '_', len);
-	fname[len] = '\0';
+	genrandname(NULL, 0);
 
 	if (tmpdir) l = strcatdelim2(path, tmpdir, NULL) - path;
+#ifdef	FAKEUNINIT
+	else l = 0;	/* fake for -Wuninitialized */
+#endif
 
 	for (;;) {
+		genrandname(fname, len);
+#if	!MSDOS
+		if (dos) {
+			int i;
+
+			for (i = 0; fname[i]; i++)
+				fname[i] = toupper2(fname[i]);
+		}
+#endif
 		if (tmpdir) {
-			strcpy(path + l, fname);
+			strcpy(&(path[l]), fname);
 			if (!isexist(path)) {
 				if (old) {
 #if	!MSDOS
-					if (isexist(fname));
+					if (isexist(fname)) /*EMPTY*/;
 					else
 #endif
-					{
-						if (rename2(old, fname) >= 0)
-							return(fname);
+					if (saferename(old, fname) >= 0)
+						return(fname);
 #if	MSDOS
-						if (errno != EACCES)
+					else if (errno == EACCES) /*EMPTY*/;
 #endif
-						break;
-					}
+					else break;
 				}
 				else {
 					if ((fd = Xopen(fname,
@@ -964,56 +1277,53 @@ char *tmpdir, *old;
 			}
 		}
 		else {
-			if (_Xmkdir(fname, 0777) >= 0) return(fname);
+			if (Xmkdir(fname, 0777) >= 0) return(fname);
 			if (errno != EEXIST) break;
 		}
-
-		for (i = len - 1; i >= 0; i--) {
-			if (fname[i] == '9') {
-				fname[i] = '_';
-				continue;
-			}
-			else if (fname[i] == '_') fname[i] = 'a';
-			else if (dos) {
-				if (fname[i] == 'z') fname[i] = '0';
-				else fname[i]++;
-			}
-			else {
-				if (fname[i] == 'z') fname[i] = 'A';
-				else if (fname[i] == 'Z') fname[i] = '0';
-				else fname[i]++;
-			}
-			break;
-		}
-		if (i < 0) break;
 	}
 	free(fname);
 	return(NULL);
 }
 
 #if	!MSDOS
-static char *NEAR getentnum(dir, bsiz, fs)
+static off_t NEAR getdirblocksize(dir)
 char *dir;
-int bsiz, fs;
 {
-	FILE *fp;
 	struct stat st;
-	char *tmp;
-	int i, n;
 
-	if (fs != 2) return(NULL);	/* without IRIX File System */
+	if (Xlstat(dir, &st) < 0) return(getblocksize(dir));
+	return((off_t)st.st_size);
+}
 
-	if (_Xlstat(dir, &st) < 0) error(dir);
-	n = (long)(st.st_size) / bsiz;
-	tmp = malloc2(n + 1);
+static u_char *NEAR getentnum(dir, bsiz)
+char *dir;
+off_t bsiz;
+{
+	struct stat st;
+	u_char ch, *tmp;
+	int i, c, n, fd;
+
+	if (Xlstat(dir, &st) < 0
+	|| (fd = Xopen(dir, O_BINARY | O_RDONLY, 0666)) < 0)
+		return(NULL);
+	n = (off_t)(st.st_size) / bsiz;
+	tmp = (u_char *)malloc2(n + 1);
 
-	fp = fopen(dir, "r");
 	for (i = 0; i < n; i++) {
-		if (fseek(fp, i * bsiz + 3, 0) < 0) error(dir);
-		tmp[i] = fgetc(fp) + 1;
+		if (Xlseek(fd, (off_t)(i * bsiz + 3), L_SET) < 0) {
+			free(tmp);
+			return(NULL);
+		}
+		while ((c = Xread(fd, (char *)&ch, sizeof(ch))) < 0
+		&& errno == EINTR);
+		if (c < 0) {
+			free(tmp);
+			return(NULL);
+		}
+		tmp[i] = ch + 1;
 	}
 	tmp[i] = 0;
-	fclose(fp);
+	Xclose(fd);
 
 	return(tmp);
 }
@@ -1026,32 +1336,36 @@ int fnamp;
 	DIR *dirp;
 	struct dirent *dp;
 
-	if (!(dirp = _Xopendir(dir))) error(dir);
-	while ((dp = Xreaddir(dirp))) {
-		if (isdotdir(dp -> d_name)) continue;
-		else {
-			strcpy(path + fnamp, dp -> d_name);
-			if (rename2(path, dp -> d_name) < 0) error(path);
+	if (!(dirp = Xopendir(dir))) warning(-1, dir);
+	else {
+		while ((dp = Xreaddir(dirp))) {
+			if (isdotdir(dp -> d_name)) continue;
+			else {
+				strcpy(&(path[fnamp]), dp -> d_name);
+				if (saferename(path, dp -> d_name) < 0)
+					warning(-1, path);
+			}
 		}
+		Xclosedir(dirp);
 	}
-	Xclosedir(dirp);
-	if (_Xrmdir(dir) < 0) error(dir);
+	if (Xrmdir(dir) < 0) warning(-1, dir);
 	free(dir);
 }
 
 VOID arrangedir(fs)
 int fs;
 {
+#if	!MSDOS
+	off_t persec, totalent, dirblocksize;
+	u_char *entnum;
+	char **tmpfiles;
+	int n, tmpno, block, ptr, totalptr;
+#endif
 	DIR *dirp;
 	struct dirent *dp;
-	int dos, headbyte, boundary, dirsize;
-	int i, top, len, fnamp, ent;
-	char *cp, *tmpdir, path[MAXPATHLEN];
-#if	!MSDOS
-	long persec;
-	char *entnum, **tmpfile;
-	int tmpno, block, totalent, ptr, totalptr;
-#endif
+	int dos, headbyte, boundary, dirsize, namofs;
+	int i, top, size, len, fnamp, ent;
+	char *cp, *tmpdir, **fnamelist, path[MAXPATHLEN];
 
 	switch (fs) {
 #if	!MSDOS
@@ -1060,12 +1374,21 @@ int fs;
 			headbyte = 4;
 			boundary = 2;
 			dirsize = sizeof(u_long);
+			namofs = 0;
 			break;
 		case 3:	/* SystemV R3 File System */
 			dos = 0;
 			headbyte = 0;
 			boundary = 8;
 			dirsize = sizeof(u_short);
+			namofs = 0;
+			break;
+		case 6:	/* Linux File System */
+			dos = 0;
+			headbyte = 0;
+			boundary = 4;
+			dirsize = 4;	/* short + short */
+			namofs = 3;
 			break;
 #endif
 		case 4:	/* MS-DOS File System */
@@ -1073,37 +1396,62 @@ int fs;
 			headbyte = -1;
 			boundary = 1;
 			dirsize = DOSDIRENT;
+			namofs = 0;
 			break;
 		case 5:	/* Windows95 File System */
 			dos = 2;
 			headbyte = -1;
 			boundary = LFNENTSIZ;
 			dirsize = DOSDIRENT;
+			namofs = 0;
 			break;
 		default:
 			dos = 0;
 			headbyte = 0;
 			boundary = 4;
-			dirsize = sizeof(u_long) + sizeof(u_short) * 2;
+			dirsize = 8;	/* long + short + short */
+			namofs = 0;
 			break;
 	}
 
-	for (i = 0; i < maxfile; i++) if (!isdotdir(filelist[i].name)) break;
-	if (i >= maxfile) return;
-	top = i;
-	if (dos == 1) len = 8;
-	else len = getnamlen(realdirsiz(filelist[top].name));
+	top = -1;
+	fnamelist = (char **)malloc2((maxfile + 1) * sizeof(char *));
+	for (i = 0; i < maxfile; i++) {
+		if (isdotdir(filelist[i].name)) cp = filelist[i].name;
+		else {
+			if (top < 0) top = i;
+			cp = convput(path, filelist[i].name, 1, 0, NULL, NULL);
+		}
+#if	!MSDOS && !defined (_NODOSDRIVE)
+		if (_dospath(cp)) cp += 2;
+#endif
+		fnamelist[i] = strdup2(cp);
+	}
+	fnamelist[i] = NULL;
+	if (top < 0) {
+		freevar(fnamelist);
+		return;
+	}
+
+	noconv++;
+	size = realdirsiz(fnamelist[top], dos, boundary, dirsize, namofs);
+	len = getnamlen(size, dos, boundary, dirsize, namofs);
 	if (!(tmpdir = maketmpfile(len, dos, NULL, NULL))) {
 		warning(0, NOWRT_K);
+		freevar(fnamelist);
+		noconv--;
 		return;
 	}
 #if	!MSDOS
 	persec = getblocksize(tmpdir);
+	dirblocksize = getdirblocksize(tmpdir);
 #endif
 	fnamp = strcatdelim2(path, tmpdir, NULL) - path;
 	waitmes();
 
-	if (!(dirp = _Xopendir("."))) {
+	if (!(dirp = Xopendir("."))) {
+		freevar(fnamelist);
+		noconv--;
 		lostcwd(path);
 		return;
 	}
@@ -1112,18 +1460,21 @@ int fs;
 		if (isdotdir(dp -> d_name)) continue;
 		else if (!strpathcmp(dp -> d_name, tmpdir)) {
 #if	MSDOS
-			if (!(dp -> d_alias[0])) len = 8;
+			if (!(dp -> d_alias[0])) len = DOSBODYLEN;
 #else
-			if (dos == 2 && dp -> d_reclen == DOSDIRENT) len = 8;
+			if (dos == 2 && dp -> d_reclen == DOSDIRENT)
+				len = DOSBODYLEN;
 #endif
 			ent = i;
 		}
 		else {
-			strcpy(path + fnamp, dp -> d_name);
-			if (rename2(dp -> d_name, path) < 0) {
+			strcpy(&(path[fnamp]), dp -> d_name);
+			if (saferename(dp -> d_name, path) < 0) {
 				Xclosedir(dirp);
 				warning(-1, dp -> d_name);
 				restorefile(tmpdir, path, fnamp);
+				freevar(fnamelist);
+				noconv--;
 				return;
 			}
 		}
@@ -1132,75 +1483,100 @@ int fs;
 	Xclosedir(dirp);
 
 	if (ent > 0) {
-		if (!(cp = maketmpfile(len, dos, tmpdir, tmpdir)))
-			error(tmpdir);
+		if (!(cp = maketmpfile(len, dos, tmpdir, tmpdir))) {
+			warning(-1, tmpdir);
+			restorefile(tmpdir, path, fnamp);
+			freevar(fnamelist);
+			noconv--;
+			return;
+		}
 		free(tmpdir);
 		tmpdir = cp;
-		strcatdelim2(path, tmpdir, NULL);
+		fnamp = strcatdelim2(path, tmpdir, NULL) - path;
 	}
 
 #if	!MSDOS
-	entnum = getentnum(".", persec, fs);
-	totalent = headbyte + realdirsiz(".") + realdirsiz("..")
-		+ realdirsiz(tmpdir);
+	if (fs != 2) entnum = NULL;	/* except IRIX File System */
+	else if (!(entnum = getentnum(".", persec))) {
+		warning(-1, ".");
+		restorefile(tmpdir, path, fnamp);
+		freevar(fnamelist);
+		noconv--;
+		return;
+	}
+	totalent = headbyte
+		+ realdirsiz(tmpdir, dos, boundary, dirsize, namofs)
+		+ realdirsiz(".", dos, boundary, dirsize, namofs)
+		+ realdirsiz("..", dos, boundary, dirsize, namofs);
 	block = tmpno = 0;
 	ptr = 3;
 	totalptr = 0;
 	if (entnum) totalptr = entnum[block];
-	tmpfile = NULL;
+	tmpfiles = NULL;
 #endif
 
 	for (i = 0; i < maxfile; i++) {
-		if (isdotdir(filelist[i].name) || i == top) continue;
+		if (isdotdir(fnamelist[i]) || i == top) continue;
 #if	!MSDOS
-		ent = persec - totalent;
+		ent = dirblocksize - totalent;
+		size = realdirsiz(fnamelist[i],
+			dos, boundary, dirsize, namofs);
 		switch (fs) {
 			case 2:	/* IRIX File System */
- 				if (totalptr > ptr + 1) ent -= totalptr;
- 				else ent -= ptr + 1;
+				if (totalptr > ptr + 1) ent -= totalptr;
+				else ent -= ptr + 1;
 				break;
 			case 3:	/* SystemV R3 File System */
 			case 4:	/* MS-DOS File System */
 			case 5:	/* Windows95 File System */
-				ent = realdirsiz(filelist[i].name);
+				ent = size;
 				break;
 			default:
 				break;
 		}
-		if (ent < realdirsiz(filelist[i].name)) {
-			tmpfile = b_realloc(tmpfile, tmpno, char *);
-			tmpfile[tmpno] =
-				maketmpfile(getnamlen(ent), dos, tmpdir, NULL);
-			tmpno++;
+		if (ent < size) {
+			n = getnamlen(ent, dos, boundary, dirsize, namofs);
+			if (n > 0) {
+				tmpfiles = b_realloc(tmpfiles, tmpno, char *);
+				tmpfiles[tmpno++] =
+					maketmpfile(n, dos, tmpdir, NULL);
+			}
 			ptr = 0;
 			totalent = headbyte;
 			if (entnum) totalptr = entnum[++block];
 		}
 #endif	/* !MSDOS */
-		strcpy(path + fnamp, filelist[i].name);
-		if (rename2(path, filelist[i].name) < 0) error(path);
+		strcpy(&(path[fnamp]), fnamelist[i]);
+		if (saferename(path, fnamelist[i]) < 0) {
+			warning(-1, path);
+			break;
+		}
 #if	!MSDOS
-		totalent += realdirsiz(filelist[i].name);
+		totalent += size;
 		ptr++;
 #endif
 	}
 #if	!MSDOS
 	if (entnum) free(entnum);
+	if (tmpfiles) {
+		for (i = 0; i < tmpno; i++) if (tmpfiles[i]) {
+			if (Xunlink(tmpfiles[i]) < 0) warning(-1, tmpfiles[i]);
+			free(tmpfiles[i]);
+		}
+		free(tmpfiles);
+	}
 #endif
 
-	if (!(cp = maketmpfile(len, dos, tmpdir, tmpdir))) error(tmpdir);
-	free(tmpdir);
-	tmpdir = cp;
-	strcatdelim2(path, tmpdir, filelist[top].name);
-	if (rename2(path, filelist[top].name) < 0) error(path);
+	if (!(cp = maketmpfile(len, dos, tmpdir, tmpdir))) warning(-1, tmpdir);
+	else {
+		free(tmpdir);
+		tmpdir = cp;
+	}
+	fnamp = strcatdelim2(path, tmpdir, fnamelist[top]) - path;
+	if (saferename(path, fnamelist[top]) < 0) warning(-1, path);
 	restorefile(tmpdir, path, fnamp);
 
-#if	!MSDOS
-	for (i = 0; i < tmpno; i++) if (tmpfile[i]) {
-		if (unlink(tmpfile[i]) < 0) error(tmpfile[i]);
-		free(tmpfile[i]);
-	}
-	free(tmpfile);
-#endif
+	freevar(fnamelist);
+	noconv--;
 }
 #endif	/* !_NOWRITEFS */
