@@ -32,7 +32,7 @@ extern int errno;
 #include <mntent.h>
 typedef struct mntent	mnt_t;
 #define	getmntent2(fp, mntp)	getmntent(fp)
-#define	hasmntopt2(mntp, opt)	strstr2((mntp) -> mnt_opts, opt)
+#define	hasmntopt2(mntp, opt)	strmntopt((mntp) -> mnt_opts, opt)
 #endif	/* USEMNTENTH */
 
 #ifdef	USEMNTTABH
@@ -41,7 +41,7 @@ typedef struct mntent	mnt_t;
 typedef struct mnttab	mnt_t;
 #define	setmntent		fopen
 #define	getmntent2(fp, mntp)	(getmntent(fp, mntp) ? NULL : mntp)
-#define	hasmntopt2(mntp, opt)	strstr2((mntp) -> mnt_mntopts, opt)
+#define	hasmntopt2(mntp, opt)	strmntopt((mntp) -> mnt_mntopts, opt)
 #define	endmntent		fclose
 #define	mnt_dir		mnt_mountp
 #define	mnt_fsname	mnt_special
@@ -77,7 +77,7 @@ typedef struct _mnt_t {
 } mnt_t;
 static FILE *NEAR setmntent __P_((char *, char *));
 static mnt_t *NEAR getmntent2 __P_((FILE *, mnt_t *));
-#define	hasmntopt2(mntp, opt)	strstr2((mntp) -> mnt_opts, opt)
+#define	hasmntopt2(mntp, opt)	strmntopt((mntp) -> mnt_opts, opt)
 #if	defined (USEMNTINFO) || defined (USEGETMNT)
 #define	endmntent(fp)
 #else
@@ -92,7 +92,7 @@ static int mnt_size = 0;
 typedef struct fstab	mnt_t;
 #define	setmntent(file, mode)	(FILE *)(setfsent(), NULL)
 #define	getmntent2(fp, mntp)	getfsent()
-#define	hasmntopt2(mntp, opt)	strstr2((mntp) -> fs_mntops, opt)
+#define	hasmntopt2(mntp, opt)	strmntopt((mntp) -> fs_mntops, opt)
 #define	endmntent(fp)		endfsent()
 #define	mnt_dir		fs_file
 #define	mnt_fsname	fs_spec
@@ -106,7 +106,7 @@ typedef struct _mnt_t {
 	char *mnt_type;
 	char *mnt_opts;
 } mnt_t;
-#define	hasmntopt2(mntp, opt)	strstr2((mntp) -> mnt_opts, opt)
+#define	hasmntopt2(mntp, opt)	strmntopt((mntp) -> mnt_opts, opt)
 # ifdef	PC98
 # define	PT_FAT12	0x81	/* 0x80 | 0x01 */
 # define	PT_FAT16	0x91	/* 0x80 | 0x11 */
@@ -124,7 +124,11 @@ typedef struct _mnt_t {
 
 #ifdef	USESTATVFSH
 #include <sys/statvfs.h>
+# ifdef	USESTATVFS_T
+typedef statvfs_t	statfs_t;
+# else
 typedef struct statvfs	statfs_t;
+# endif
 #define	statfs2		statvfs
 #define	blocksize(fs)	((fs).f_frsize ? (fs).f_frsize : (fs).f_bsize)
 #endif	/* USESTATVFSH */
@@ -155,6 +159,9 @@ typedef struct fs_data	statfs_t;
 #define	blocksize(fs)		1024
 #define	f_bsize		fd_req.bsize
 #define	f_files		fd_req.gtot
+#define	f_blocks	fd_req.btot
+#define	f_bfree		fd_req.bfree
+#define	f_bavail	fd_req.bfreen
 #endif	/* USEFSDATA */
 
 #ifdef	USEFFSIZE
@@ -189,6 +196,7 @@ extern int unixstatfs __P_((char *, statfs_t *));
 extern VOID error __P_((char *));
 extern int _chdir2 __P_((char *));
 extern char *strcpy2 __P_((char *, char *));
+extern int snprintf2 __P_((char *, int, CONST char *, ...));
 extern char *getwd2 __P_((VOID_A));
 extern VOID warning __P_((int, char *));
 #if	MSDOS || !defined (_NODOSDRIVE)
@@ -211,8 +219,7 @@ extern int dosstatfs __P_((int, char *));
 #endif
 extern char *malloc2 __P_((ALLOC_T));
 extern char *realloc2 __P_((VOID_P, ALLOC_T));
-extern char *strstr2 __P_((char *, char *));
-extern char *ascnumeric __P_((char *, long, int, int));
+extern char *ascnumeric __P_((char *, off_t, int, int));
 extern int kanjiputs __P_((char *));
 extern int kanjiputs2 __P_((char *, int, int));
 extern int Xaccess __P_((char *, int));
@@ -227,6 +234,8 @@ extern char *distributor;
 extern int needbavail;
 #endif
 
+#define	KEYWID		7
+
 #ifndef	MOUNTED
 #define	MOUNTED		"/etc/mtab"
 #endif
@@ -243,6 +252,15 @@ extern int needbavail;
 #ifndef	MNTTYPE_FFS
 #define	MNTTYPE_FFS	"ffs"	/* NetBSD, OpenBSD */
 #endif
+#ifndef	MNTTYPE_ADVFS
+#define	MNTTYPE_ADVFS	"advfs"	/* OSF/1 */
+#endif
+#ifndef	MNTTYPE_VXFS
+#define	MNTTYPE_VXFS	"vxfs"	/* HP-UX */
+#endif
+#ifndef	MNTTYPE_HFS
+#define	MNTTYPE_HFS	"hfs"	/* Darwin, HP-UX */
+#endif
 #ifndef	MNTTYPE_EXT2
 #define	MNTTYPE_EXT2	"ext2"	/* Linux */
 #endif
@@ -258,6 +276,15 @@ extern int needbavail;
 #ifndef	MNTTYPE_DGUX
 #define	MNTTYPE_DGUX	"dg/ux"	/* DG/UX */
 #endif
+#ifndef	MNTTYPE_MSDOS
+#define	MNTTYPE_MSDOS	"msdos"	/* msdosfs */
+#endif
+#ifndef	MNTTYPE_UMSDOS
+#define	MNTTYPE_UMSDOS	"umsdos"	/* umsdosfs */
+#endif
+#ifndef	MNTTYPE_VFAT
+#define	MNTTYPE_VFAT	"vfat"	/* vfatfs */
+#endif
 #ifndef	MNTTYPE_PC
 #define	MNTTYPE_PC	"pc"	/* MS-DOS */
 #endif
@@ -281,13 +308,12 @@ static int NEAR code2str __P_((char *, int));
 static int NEAR checkline __P_((int));
 VOID help __P_((int));
 static int NEAR getfsinfo __P_((char *, statfs_t *, mnt_t *));
+static char *NEAR strmntopt __P_((char *, char *));
 int writablefs __P_((char *));
-long getblocksize __P_((char *));
-static int NEAR info1line __P_((int, char *, long, char *, char *));
-#if	!defined (USEFSDATA) || !defined (_NODOSDRIVE)
-static long NEAR calcKB __P_((long, long));
-#endif
-VOID getinfofs __P_((char *, long *, long *));
+off_t getblocksize __P_((char *));
+static int NEAR info1line __P_((int, char *, off_t, char *, char *));
+off_t calcKB __P_((off_t, off_t));
+int getinfofs __P_((char *, off_t *, off_t *, off_t *));
 int infofs __P_((char *));
 
 static int keycodelist[] = {
@@ -302,32 +328,40 @@ static char *keystrlist[] = {
 	"Bs", "Tab", "Ret", "Esc"
 };
 
+
 static int NEAR code2str(buf, code)
 char *buf;
 int code;
 {
 	buf = buf + strlen(buf);
 	if (code >= K_F(1) && code <= K_F(20))
-		sprintf(buf, "F%-6d", code - K_F0);
+		snprintf2(buf, KEYWID + 1, "F%-6d", code - K_F0);
 	else if ((code & ~0x7f) == 0x80 && isalpha(code & 0x7f))
-		sprintf(buf, "Alt-%c  ", code & 0x7f);
-	else if (code == K_UP) sprintf(buf, "%-7.7s", UPAR_K);
-	else if (code == K_DOWN) sprintf(buf, "%-7.7s", DWNAR_K);
-	else if (code == K_RIGHT) sprintf(buf, "%-7.7s", RIGAR_K);
-	else if (code == K_LEFT) sprintf(buf, "%-7.7s", LEFAR_K);
+		snprintf2(buf, KEYWID + 1, "Alt-%c  ", code & 0x7f);
+	else if (code == K_UP)
+		snprintf2(buf, KEYWID + 1, "%-7.7s", UPAR_K);
+	else if (code == K_DOWN)
+		snprintf2(buf, KEYWID + 1, "%-7.7s", DWNAR_K);
+	else if (code == K_RIGHT)
+		snprintf2(buf, KEYWID + 1, "%-7.7s", RIGAR_K);
+	else if (code == K_LEFT)
+		snprintf2(buf, KEYWID + 1, "%-7.7s", LEFAR_K);
 	else {
 		int i;
 
 		for (i = 0; i < KEYCODESIZ; i++)
 			if (code == keycodelist[i]) break;
-		if (i < KEYCODESIZ) sprintf(buf, "%-7.7s", keystrlist[i]);
+		if (i < KEYCODESIZ)
+			snprintf2(buf, KEYWID + 1, "%-7.7s", keystrlist[i]);
 #ifndef	CODEEUC
-		else if (iskna(code)) sprintf(buf, "'%c'    ", code);
+		else if (iskna(code))
+			snprintf2(buf, KEYWID + 1, "'%c'    ", code);
 #endif
 		else if (isctl(code))
-			sprintf(buf, "Ctrl-%c ", (code + '@') & 0x7f);
+			snprintf2(buf, KEYWID + 1,
+				"Ctrl-%c ", (code + '@') & 0x7f);
 		else if (code < K_MIN && !ismsb(code))
-			sprintf(buf, "'%c'    ", code);
+			snprintf2(buf, KEYWID + 1, "'%c'    ", code);
 		else return(0);
 	}
 	return(1);
@@ -351,7 +385,7 @@ int mode;
 
 	if (distributor && *distributor) {
 		i = n_column - (int)strlen(distributor) - 24;
-		locate(i, LHELP);
+		locate(i, L_HELP);
 		putch2('[');
 		putterm(t_standout);
 		cprintf2(" Distributed by: %s ", distributor);
@@ -369,10 +403,10 @@ int mode;
 		locate(x * (n_column / 2), y);
 		if (x ^= 1) putterm(l_clear);
 		else y++;
-		if (mode && !(funclist[i].stat & ARCH)) continue;
+		if (mode && !(funclist[i].status & ARCH)) continue;
 
 		c = 0;
-		*buf = '\0';
+		buf[0] = '\0';
 		for (j = 0; j < MAXBINDTABLE && bindlist[j].key >= 0; j++)
 			if (i == (int)(bindlist[j].f_func)) {
 				if ((c += code2str(buf,
@@ -425,7 +459,6 @@ mnt_t *mntp;
 	static char *fsname = NULL;
 	static char *dir = NULL;
 	static char *type = NULL;
-	static char *opts = NULL;
 	struct vfs_ent *entp;
 	struct vmount *vmntp;
 	char *cp, *buf, *host;
@@ -469,7 +502,7 @@ mnt_t *mntp;
 	mntp -> mnt_fsname = fsname;
 	mntp -> mnt_dir = dir;
 	mntp -> mnt_type = (type) ? type : "???";
-	mntp -> mnt_opts = (opts) ? opts : "";
+	mntp -> mnt_opts = (vmntp -> vmt_flags & MNT_READONLY) ? "ro" : "";
 
 	mnt_ptr += vmntp -> vmt_length;
 	return(mntp);
@@ -481,15 +514,18 @@ mnt_t *mntp;
 static FILE *NEAR setmntent(file, mode)
 char *file, *mode;
 {
-	struct statfs *buf;
+#ifndef	USEMNTINFO
 	int size;
+#endif
+	struct statfs *buf;
 
 	buf = NULL;
-	mnt_ptr = mnt_size = size = 0;
+	mnt_ptr = mnt_size = 0;
 #ifdef	USEMNTINFO
 	mnt_size = getmntinfo(&buf, MNT_NOWAIT);
 #else
 # ifdef	USEMNTINFOR
+	size = 0;
 	getmntinfo_r(&buf, MNT_WAIT, &mnt_size, &size);
 # else
 	size = (getfsstat(NULL, 0, MNT_WAIT) + 1) * sizeof(struct statfs);
@@ -675,7 +711,10 @@ mnt_t *mntbuf;
 		memset((char *)fsbuf, 0xff, sizeof(statfs_t));
 	}
 #else	/* !MSDOS */
-	mnt_t *mntp, mnt;
+# if	!defined (USEMNTENTH) && !defined (USEGETFSENT)
+	mnt_t mnt;
+# endif
+	mnt_t *mntp;
 	FILE *fp;
 	char *dir, fsname[MAXPATHLEN];
 	int len, match;
@@ -703,20 +742,20 @@ mnt_t *mntbuf;
 		if (buf[sizeof(long) * 3] & 001)
 			mntbuf -> mnt_type = MNTTYPE_FAT32;
 		fsbuf -> f_bsize = *((long *)&(buf[sizeof(long) * 0]));
-#ifdef	USEFSDATA
-		fsbuf -> fd_req.btot = calcKB(fsbuf -> f_bsize,
-			*((long *)&(buf[sizeof(long) * 1])));
+# ifdef	USEFSDATA
+		fsbuf -> fd_req.btot = calcKB((off_t)(fsbuf -> f_bsize),
+			(off_t)(*((long *)&(buf[sizeof(long) * 1]))));
 		fsbuf -> fd_req.bfree =
-		fsbuf -> fd_req.bfreen = calcKB(fsbuf -> f_bsize,
-			*((long *)&(buf[sizeof(long) * 2])));
-#else	/* !USEFSDATA */
-#ifdef	USESTATVFSH
+		fsbuf -> fd_req.bfreen = calcKB((off_t)(fsbuf -> f_bsize),
+			(off_t)(*((long *)&(buf[sizeof(long) * 2]))));
+# else	/* !USEFSDATA */
+#  ifdef	USESTATVFSH
 		fsbuf -> f_frsize = 0;
-#endif
+#  endif
 		fsbuf -> f_blocks = *((long *)&(buf[sizeof(long) * 1]));
 		fsbuf -> f_bfree =
 		fsbuf -> f_bavail = *((long *)&(buf[sizeof(long) * 2]));
-#endif	/* !USEFSDATA */
+# endif	/* !USEFSDATA */
 		fsbuf -> f_files = -1;
 		return(0);
 	}
@@ -775,6 +814,21 @@ mnt_t *mntbuf;
 	return(0);
 }
 
+static char *NEAR strmntopt(s1, s2)
+char *s1, *s2;
+{
+	char *cp;
+	int len;
+
+	len = strlen(s2);
+	for (cp = s1; cp && *cp;) {
+		if (!strncmp(cp, s2, len) && (!cp[len] || cp[len] == ','))
+			return(cp);
+		if ((cp = strchr(cp, ','))) cp++;
+	}
+	return(NULL);
+}
+
 int writablefs(path)
 char *path;
 {
@@ -792,33 +846,46 @@ char *path;
 	if (getfsinfo(path, &fsbuf, &mntbuf) < 0
 	|| hasmntopt2(&mntbuf, "ro")) return(0);
 
+	if (!strcmp(mntbuf.mnt_type, MNTTYPE_PC)) return(4);
+	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_DOS7)
+	|| !strcmp(mntbuf.mnt_type, MNTTYPE_FAT32)) return(5);
 #if	!MSDOS
-	if (!strcmp(mntbuf.mnt_type, MNTTYPE_43)
+	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_43)
 	|| !strcmp(mntbuf.mnt_type, MNTTYPE_42)
 	|| !strcmp(mntbuf.mnt_type, MNTTYPE_UFS)
 	|| !strcmp(mntbuf.mnt_type, MNTTYPE_FFS)
-	|| !strcmp(mntbuf.mnt_type, MNTTYPE_EXT2)
 	|| !strcmp(mntbuf.mnt_type, MNTTYPE_JFS)) return(1);
 	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_EFS)) return(2);
 	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_SYSV)
 	|| !strcmp(mntbuf.mnt_type, MNTTYPE_DGUX)) return(3);
-	else
+	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_EXT2)) return(6);
+	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_MSDOS))
+# ifdef	LINUX
+		return(4);
+# else
+		return(5);
+# endif
+	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_UMSDOS)
+	|| !strcmp(mntbuf.mnt_type, MNTTYPE_VFAT)) return(5);
+	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_ADVFS)) return(0);
+	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_VXFS)) return(0);
+#endif
+#ifdef	DARWIN
+	/* Macintosh HFS+ is pseudo file system covered with skin */
+	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_HFS)) return(0);
 #endif
-	if (!strcmp(mntbuf.mnt_type, MNTTYPE_PC)) return(4);
-	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_DOS7)) return(5);
-	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_FAT32)) return(5);
 	return(0);
 }
 
 /*ARGSUSED*/
-long getblocksize(dir)
+off_t getblocksize(dir)
 char *dir;
 {
 #if	MSDOS
 	statfs_t fsbuf;
 
-	if (statfs2(dir, &fsbuf) < 0) return(1024);
-	return(blocksize(fsbuf));
+	if (statfs2(dir, &fsbuf) < 0) return((off_t)1024);
+	return((off_t)blocksize(fsbuf));
 #else	/* !MSDOS */
 	statfs_t fsbuf;
 	mnt_t mntbuf;
@@ -827,12 +894,12 @@ char *dir;
 #endif
 
 	if (!strcmp(dir, ".") && getfsinfo(dir, &fsbuf, &mntbuf) >= 0)
-		return(blocksize(fsbuf));
+		return((off_t)blocksize(fsbuf));
 #ifdef	DEV_BSIZE
 	return(DEV_BSIZE);
 #else
 	if (Xstat(dir, &st) < 0) error(dir);
-	return((int)(st.st_size));
+	return((off_t)(st.st_size));
 #endif
 #endif	/* !MSDOS */
 }
@@ -840,7 +907,7 @@ char *dir;
 static int NEAR info1line(y, ind, n, s, unit)
 int y;
 char *ind;
-long n;
+off_t n;
 char *s, *unit;
 {
 	char buf[MAXCOLSCOMMA(3) + 1];
@@ -862,25 +929,24 @@ char *s, *unit;
 	return(checkline(++y));
 }
 
-#if	!defined (USEFSDATA) || !defined (_NODOSDRIVE)
-static long NEAR calcKB(block, byte)
-long block, byte;
+off_t calcKB(block, byte)
+off_t block, byte;
 {
-	if (block < 0 || byte <= 0) return(-1);
-	if (byte >= 1024) {
-		byte = (byte + 512) / 1024;
+	if (block < (off_t)0 || byte <= (off_t)0) return((off_t)-1);
+	if (byte == (off_t)1024) return(block);
+	else if (byte > 1024) {
+		byte = (byte + (off_t)512L) / (off_t)1024;
 		return(block * byte);
 	}
 	else {
-		byte = (1024 + (byte / 2)) / byte;
+		byte = ((off_t)1024 + (byte / (off_t)2)) / byte;
 		return(block / byte);
 	}
 }
-#endif
 
-VOID getinfofs(path, totalp, freep)
+int getinfofs(path, totalp, freep, bsizep)
 char *path;
-long *totalp, *freep;
+off_t *totalp, *freep, *bsizep;
 {
 	statfs_t fsbuf;
 	mnt_t mntbuf;
@@ -888,20 +954,14 @@ long *totalp, *freep;
 #ifndef	_NODOSDRIVE
 	needbavail = 1;
 #endif
-	if (getfsinfo(path, &fsbuf, &mntbuf) < 0)
-		*totalp = *freep = -1;
-	else {
-#ifdef	USEFSDATA
-		*totalp = fsbuf.fd_req.btot;
-		*freep = fsbuf.fd_req.bfreen;
-#else
-		*totalp = calcKB(fsbuf.f_blocks, blocksize(fsbuf));
-		*freep = calcKB(fsbuf.f_bavail, blocksize(fsbuf));
-#endif
-	}
+	if (getfsinfo(path, &fsbuf, &mntbuf) < 0) return(-1);
+	*totalp = fsbuf.f_blocks;
+	*freep = fsbuf.f_bavail;
+	*bsizep = blocksize(fsbuf);
 #ifndef	_NODOSDRIVE
 	needbavail = 0;
 #endif
+	return(0);
 }
 
 int infofs(path)
@@ -925,25 +985,21 @@ char *path;
 	locate(0, y++);
 	putterm(l_clear);
 
-	y = info1line(y, FSNAM_K, 0L, mntbuf.mnt_fsname, NULL);
-	y = info1line(y, FSMNT_K, 0L, mntbuf.mnt_dir, NULL);
-	y = info1line(y, FSTYP_K, 0L, mntbuf.mnt_type, NULL);
-#ifdef	USEFSDATA
-	y = info1line(y, FSTTL_K, fsbuf.fd_req.btot, NULL, "Kbytes");
-	y = info1line(y, FSUSE_K,
-		fsbuf.fd_req.btot - fsbuf.fd_req.bfree, NULL, "Kbytes");
-	y = info1line(y, FSAVL_K, fsbuf.fd_req.bfreen, NULL, "Kbytes");
-#else
+	y = info1line(y, FSNAM_K, (off_t)0, mntbuf.mnt_fsname, NULL);
+	y = info1line(y, FSMNT_K, (off_t)0, mntbuf.mnt_dir, NULL);
+	y = info1line(y, FSTYP_K, (off_t)0, mntbuf.mnt_type, NULL);
 	y = info1line(y, FSTTL_K,
-		calcKB(fsbuf.f_blocks, blocksize(fsbuf)), NULL, "Kbytes");
+		calcKB((off_t)(fsbuf.f_blocks), (off_t)blocksize(fsbuf)),
+		NULL, "Kbytes");
 	y = info1line(y, FSUSE_K,
-		calcKB(fsbuf.f_blocks - fsbuf.f_bfree, blocksize(fsbuf)),
+		calcKB((off_t)(fsbuf.f_blocks - fsbuf.f_bfree),
+			(off_t)blocksize(fsbuf)),
 		NULL, "Kbytes");
 	y = info1line(y, FSAVL_K,
-		calcKB(fsbuf.f_bavail, blocksize(fsbuf)), NULL, "Kbytes");
-#endif
-	y = info1line(y, FSBSZ_K, fsbuf.f_bsize, NULL, "bytes");
-	y = info1line(y, FSINO_K, fsbuf.f_files, NULL, UNIT_K);
+		calcKB((off_t)(fsbuf.f_bavail), (off_t)blocksize(fsbuf)),
+		NULL, "Kbytes");
+	y = info1line(y, FSBSZ_K, (off_t)(fsbuf.f_bsize), NULL, "bytes");
+	y = info1line(y, FSINO_K, (off_t)(fsbuf.f_files), NULL, UNIT_K);
 	if (y > LFILETOP + 1) {
 		for (; y < LFILEBOTTOM; y++) {
 			locate(0, y);
