@@ -6,15 +6,14 @@
 
 #include <stdio.h>
 #include <string.h>
-#include <sys/types.h>
 #include <fcntl.h>
 #include "machine.h"
 
-#ifndef NOUNISTDH
+#ifndef	NOUNISTDH
 #include <unistd.h>
 #endif
 
-#ifndef NOSTDLIBH
+#ifndef	NOSTDLIBH
 #include <stdlib.h>
 #endif
 
@@ -26,12 +25,6 @@
 #include "kctype.h"
 #include "pathname.h"
 
-#ifdef	USESTDARGH
-#include <stdarg.h>
-#else
-#include <varargs.h>
-#endif
-
 #if	MSDOS
 #include <io.h>
 #include "unixemu.h"
@@ -40,6 +33,14 @@
 #include <sys/param.h>
 #endif
 
+#if	MSDOS && defined (_NOUSELFN) && !defined (_NODOSDRIVE)
+#define	_NODOSDRIVE
+#endif
+
+#if	defined (_NOKANJICONV) && !defined (_NOKANJIFCONV)
+#define	_NOKANJIFCONV
+#endif
+
 #ifndef	L_SET
 # ifdef	SEEK_SET
 # define	L_SET	SEEK_SET
@@ -52,7 +53,6 @@
 #define	KANA		001
 #define	KANJI		002
 #define	JKANA		004
-#define	MAXPRINTBUF	1023
 #define	SJ_UDEF		0x81ac	/* GETA */
 #define	UNICODETBL	"fd-unicd.tbl"
 #define	MINUNICODE	0x00a7
@@ -74,8 +74,14 @@ typedef struct _kconv_t {
 extern char *malloc2 __P_((ALLOC_T));
 extern char *realloc2 __P_((VOID_P, ALLOC_T));
 extern char *strncpy3 __P_((char *, char *, int *, int));
-extern char *strstr2 __P_((char *, char *));
+extern int strlen3 __P_((char *));
+extern VOID error __P_((char *));
+extern off_t Xlseek __P_((int, off_t, int));
 
+#if	(!MSDOS && !defined (_NOKANJICONV)) \
+|| (!defined (_NOENGMES) && !defined (_NOJPNMES))
+static char *NEAR strstr2 __P_((char *, char *));
+#endif
 #if	!MSDOS && (!defined (_NOKANJICONV) \
 || (!defined (_NODOSDRIVE) && defined (CODEEUC)))
 static VOID NEAR sj2j __P_((char *, u_char *));
@@ -84,12 +90,10 @@ static VOID NEAR j2sj __P_((char *, u_char *));
 #if	!MSDOS && !defined (_NOKANJICONV)
 static int NEAR toenglish __P_((char *, u_char *, int));
 static int NEAR tojis7 __P_((char *, u_char *, int, int, int, int));
-static int NEAR fromjis7 __P_((char *, u_char *, int, int, int, int));
+static int NEAR fromjis __P_((char *, u_char *, int, int));
 # if	defined (FD) && (FD >= 2)
 static int NEAR tojis8 __P_((char *, u_char *, int, int, int, int));
-static int NEAR fromjis8 __P_((char *, u_char *, int, int, int, int));
 static int NEAR tojunet __P_((char *, u_char *, int, int, int, int));
-static int NEAR fromjunet __P_((char *, u_char *, int, int, int, int));
 static int NEAR toutf8 __P_((char *, u_char *, int));
 static int NEAR fromutf8 __P_((char *, u_char *, int));
 #  ifndef	_NOKANJIFCONV
@@ -114,6 +118,9 @@ int outputkcode = 0;
 #if	!MSDOS && !defined (_NOKANJIFCONV)
 int fnamekcode = 0;
 #endif
+#if	(!MSDOS && !defined (_NOKANJICONV)) || !defined (_NODOSDRIVE)
+char *unitblpath = NULL;
+#endif
 #ifndef	LSI_C
 u_char kctypetable[256] = {
 	0200, 0200, 0200, 0200, 0200, 0200, 0200, 0200,	/* 0x00 */
@@ -150,62 +157,66 @@ u_char kctypetable[256] = {
 	0013, 0013, 0013, 0013, 0013, 0010, 0010,    0
 };
 #endif	/* !LSI_C */
+
 #if	(!MSDOS && !defined (_NOKANJICONV)) || !defined (_NODOSDRIVE)
-char *unitblpath = NULL;
 static u_char *unitblbuf = NULL;
-static u_short unitblent = 0;
+static u_int unitblent = 0;
 static kconv_t rsjistable[] = {
-	{0x8470, 0x8440, 0x0f},		/* strange Russian char */
+	{0x8470, 0x8440, 0x0f},		/* Cyrillic small letters */
 	{0x8480, 0x844f, 0x12},		/* Why they converted ? */
 #define	EXCEPTRUSS	2
-	{0x8754, 0xfa4a, 0x0a},
-	{0x8782, 0xfa59, 0x01},
-	{0x8784, 0xfa5a, 0x01},
-	{0x878a, 0xfa58, 0x01},
-	{0x8790, 0x81e0, 0x01},
-	{0x8791, 0x81df, 0x01},
-	{0x8792, 0x81e7, 0x01},
-	{0x8795, 0x81e3, 0x01},
-	{0x8796, 0x81db, 0x01},
-	{0x8797, 0x81da, 0x01},
-	{0x879a, 0x81e6, 0x01},
-	{0x879b, 0x81bf, 0x01},
-	{0x879c, 0x81be, 0x01},
-	{0xed40, 0xfa5c, 0x23},
-	{0xed63, 0xfa80, 0x1c},
-	{0xed80, 0xfa9c, 0x61},
-	{0xede1, 0xfb40, 0x1c},
-	{0xee40, 0xfb5c, 0x23},
-	{0xee63, 0xfb80, 0x1c},
-	{0xee80, 0xfb9c, 0x61},
-	{0xeee1, 0xfc40, 0x0c},
-	{0xeeef, 0xfa40, 0x0a},
-	{0xeef9, 0x81ca, 0x01},
-	{0xeefa, 0xfa55, 0x03},
-	{0xfa54, 0x81ca, 0x01},
-	{0xfa5b, 0x81e6, 0x01},
+	{0x8754, 0xfa4a, 0x0a},		/* Roman numerals */
+	{0x8782, 0xfa59, 0x01},		/* numero sign */
+	{0x8784, 0xfa5a, 0x01},		/* telephone sign */
+	{0x878a, 0xfa58, 0x01},		/* parenthesized ideograph stock */
+	{0x8790, 0x81e0, 0x01},		/* nearly equals */
+	{0x8791, 0x81df, 0x01},		/* identical to */
+	{0x8792, 0x81e7, 0x01},		/* integral */
+	{0x8795, 0x81e3, 0x01},		/* square root */
+	{0x8796, 0x81db, 0x01},		/* up tack */
+	{0x8797, 0x81da, 0x01},		/* angle */
+	{0x879a, 0x81e6, 0x01},		/* because */
+	{0x879b, 0x81bf, 0x01},		/* intersection */
+	{0x879c, 0x81be, 0x01},		/* union */
+	{0xed40, 0xfa5c, 0x23},		/* NEC-selected IBM extensions */
+	{0xed63, 0xfa80, 0x1c},		/* NEC-selected IBM extensions */
+	{0xed80, 0xfa9c, 0x61},		/* NEC-selected IBM extensions */
+	{0xede1, 0xfb40, 0x1c},		/* NEC-selected IBM extensions */
+	{0xee40, 0xfb5c, 0x23},		/* NEC-selected IBM extensions */
+	{0xee63, 0xfb80, 0x1c},		/* NEC-selected IBM extensions */
+	{0xee80, 0xfb9c, 0x61},		/* NEC-selected IBM extensions */
+	{0xeee1, 0xfc40, 0x0c},		/* NEC-selected IBM extensions */
+	{0xeeef, 0xfa40, 0x0a},		/* small Roman numerals */
+	{0xeef9, 0x81ca, 0x01},		/* full width not sign */
+	{0xeefa, 0xfa55, 0x03},		/* full width broken bar */
+					/* full width apostrophe */
+					/* full width quotation mark */
+	{0xfa54, 0x81ca, 0x01},		/* full width not sign */
+	{0xfa5b, 0x81e6, 0x01},		/* because */
 };
 #define	RSJISTBLSIZ	((int)(sizeof(rsjistable) / sizeof(kconv_t)))
 #endif	/* (!MSDOS && !_NOKANJICONV) || !_NODOSDRIVE */
 #if	!MSDOS && (!defined (_NOKANJICONV) \
 || (!defined (_NODOSDRIVE) && defined (CODEEUC)))
 static kconv_t convtable[] = {
-	{0xfa40, 0xeeef, 0x0a},
-	{0xfa4a, 0x8754, 0x0a},
-	{0xfa54, 0x81ca, 0x01},
-	{0xfa55, 0xeefa, 0x03},
-	{0xfa58, 0x878a, 0x01},
-	{0xfa59, 0x8782, 0x01},
-	{0xfa5a, 0x8784, 0x01},
-	{0xfa5b, 0x81e6, 0x01},
-	{0xfa5c, 0xed40, 0x23},
-	{0xfa80, 0xed63, 0x1c},
-	{0xfa9c, 0xed80, 0x61},
-	{0xfb40, 0xede1, 0x1c},
-	{0xfb5c, 0xee40, 0x23},
-	{0xfb80, 0xee63, 0x1c},
-	{0xfb9c, 0xee80, 0x61},
-	{0xfc40, 0xeee1, 0x0c}
+	{0xfa40, 0xeeef, 0x0a},		/* small Roman numerals */
+	{0xfa4a, 0x8754, 0x0a},		/* Roman numerals */
+	{0xfa54, 0x81ca, 0x01},		/* full width not sign */
+	{0xfa55, 0xeefa, 0x03},		/* full width broken bar */
+					/* full width apostrophe */
+					/* full width quotation mark */
+	{0xfa58, 0x878a, 0x01},		/* parenthesized ideograph stock */
+	{0xfa59, 0x8782, 0x01},		/* numero sign */
+	{0xfa5a, 0x8784, 0x01},		/* telephone sign */
+	{0xfa5b, 0x81e6, 0x01},		/* because */
+	{0xfa5c, 0xed40, 0x23},		/* IBM extensions */
+	{0xfa80, 0xed63, 0x1c},		/* IBM extensions */
+	{0xfa9c, 0xed80, 0x61},		/* IBM extensions */
+	{0xfb40, 0xede1, 0x1c},		/* IBM extensions */
+	{0xfb5c, 0xee40, 0x23},		/* IBM extensions */
+	{0xfb80, 0xee63, 0x1c},		/* IBM extensions */
+	{0xfb9c, 0xee80, 0x61},		/* IBM extensions */
+	{0xfc40, 0xeee1, 0x0c}		/* IBM extensions */
 };
 #define	CNVTBLSIZ	((int)(sizeof(convtable) / sizeof(kconv_t)))
 static u_char sj2jtable1[256] = {
@@ -293,7 +304,7 @@ static u_char j2sjtable2[128] = {
 	0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e,    0
 };
 #endif	/* !MSDOS && (!_NOKANJICONV || (!_NODOSDRIVE && CODEEUC)) */
-#if	!MSDOS && !defined (_NOKANJIFCONV)
+#if	!MSDOS && !defined (_NOKANJIFCONV) && defined (FD) && (FD >= 2)
 static u_char hctypetable[256] = {
 	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	/* 0x00 */
 	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	/* 0x10 */
@@ -351,7 +362,7 @@ static u_char b2htable[256] = "0123456789abcdef";
 #define	iscap(s, i)	(((s)[i] == CAPTAG) \
 			&& (hctypetable[(u_char)(s)[(i) + 1]] & HC_CAP) \
 			&& (hctypetable[(u_char)(s)[(i) + 2]] & HC_HEX))
-#endif	/* !MSDOS && !_NOKANJIFCONV */
+#endif	/* !MSDOS && !_NOKANJIFCONV && FD && FD >= 2 */
 
 int onkanji1 __P_((char *, int));
 #if	(!MSDOS && !defined (_NOKANJICONV)) \
@@ -370,8 +381,8 @@ int ujis2sjis __P_((char *, u_char *, int));
 || !defined (_NODOSDRIVE)
 VOID readunitable __P_((VOID_A));
 VOID discardunitable __P_((VOID_A));
-u_short unifysjis __P_((u_short, int));
-u_short cnvunicode __P_((u_short, int));
+u_int unifysjis __P_((u_int, int));
+u_int cnvunicode __P_((u_int, int));
 #endif
 #if	!MSDOS && !defined (_NOKANJICONV)
 int kanjiconv __P_((char *, char *, int, int, int, int));
@@ -389,28 +400,52 @@ int onkanji1(s, ptr)
 char *s;
 int ptr;
 {
-	int i, j;
+	int i;
 
 	if (ptr < 0) return(0);
 	if (!ptr) return(iskanji1(s, 0));
 
-	for (i = j = 0; i < ptr; i++, j++) {
-		if (!s[j]) return(0);
+	for (i = 0; i < ptr; i++) {
+		if (!s[i]) return(0);
 #ifdef	CODEEUC
-		if (isekana(s, j)) j++;
+		if (isekana(s, i)) i++;
 		else
 #endif
-		if (iskanji1(s, j)) {
-			i++;
-			j++;
-		}
+		if (iskanji1(s, i)) i++;
 	}
 	if (i > ptr) return(0);
-	return(iskanji1(s, j));
+	return(iskanji1(s, i));
 }
 
 #if	(!MSDOS && !defined (_NOKANJICONV)) \
 || (!defined (_NOENGMES) && !defined (_NOJPNMES))
+static char *NEAR strstr2(s1, s2)
+char *s1, *s2;
+{
+	int i, c1, c2;
+
+	while (*s1) {
+		for (i = 0;; i++) {
+			if (!s2[i]) return(s1);
+			c1 = toupper(s1[i]);
+			c2 = toupper(s2[i]);
+			if (c1 != c2) break;
+#ifndef	CODEEUC
+			if (issjis1(c1)) {
+				if (!s2[++i]) return(s1);
+				if (s1[i] != s2[i]) break;
+			}
+#endif
+			if (!s1[i]) break;
+		}
+#ifndef	CODEEUC
+		if (issjis1(*s1)) s1++;
+#endif
+		s1++;
+	}
+	return(NULL);
+}
+
 /*ARGSUSED*/
 int getlang(s, io)
 char *s;
@@ -421,29 +456,24 @@ int io;
 	if (!s) ret = NOCNV;
 # if	!MSDOS && !defined (_NOKANJICONV)
 #  ifndef	_NOKANJIFCONV
-	else if (io == L_FNAME && (strstr2(s, "HEX") || strstr2(s, "hex")))
-		ret = HEX;
-	else if (io == L_FNAME && (strstr2(s, "CAP") || strstr2(s, "cap")))
-		ret = CAP;
+	else if (io == L_FNAME && strstr2(s, "hex")) ret = HEX;
+	else if (io == L_FNAME && strstr2(s, "cap")) ret = CAP;
 #  endif	/* !_NOKANJIFCONV */
-	else if (strstr2(s, "SJIS") || strstr2(s, "sjis")) ret = SJIS;
-	else if (strstr2(s, "EUC") || strstr2(s, "euc")
-	|| strstr2(s, "UJIS") || strstr2(s, "ujis")) ret = EUC;
+	else if (strstr2(s, "sjis")) ret = SJIS;
+	else if (strstr2(s, "euc") || strstr2(s, "ujis")) ret = EUC;
 #  if	defined (FD) && (FD >= 2)
-	else if (strstr2(s, "UTF8") || strstr2(s, "utf8")) ret = UTF8;
-	else if (strstr2(s, "OJUNET") || strstr2(s, "ojunet"))
-		ret = O_JUNET;
-	else if (strstr2(s, "OJIS8") || strstr2(s, "ojis8")) ret = O_JIS8;
-	else if (strstr2(s, "OJIS") || strstr2(s, "ojis")) ret = O_JIS7;
-	else if (strstr2(s, "JUNET") || strstr2(s, "junet")) ret = JUNET;
-	else if (strstr2(s, "JIS8") || strstr2(s, "jis8")) ret = JIS8;
+	else if (strstr2(s, "utf8") || strstr2(s, "utf-8")) ret = UTF8;
+	else if (strstr2(s, "ojunet")) ret = O_JUNET;
+	else if (strstr2(s, "ojis8")) ret = O_JIS8;
+	else if (strstr2(s, "ojis")) ret = O_JIS7;
+	else if (strstr2(s, "junet")) ret = JUNET;
+	else if (strstr2(s, "jis8")) ret = JIS8;
 #  endif	/* FD && (FD >= 2) */
-	else if (strstr2(s, "JIS") || strstr2(s, "jis")) ret = JIS7;
+	else if (strstr2(s, "jis")) ret = JIS7;
 # endif	/* !MSDOS && !_NOKANJICONV */
 # ifndef	_NOENGMES
-	else if (io == L_OUTPUT
-	&& (strstr2(s, "ENG") || strstr2(s, "eng")
-	|| !strcmp(s, "C"))) ret = ENG;
+	else if (io == L_OUTPUT && (strstr2(s, "eng") || !strcmp(s, "C")))
+		ret = ENG;
 # endif
 	else ret = NOCNV;
 
@@ -479,10 +509,10 @@ u_char *s;
 	s1 = s[0] & 0xff;
 	s2 = s[1] & 0xff;
 	if (s1 >= 0xf0) {
-		u_short w;
+		u_int w;
 		int i;
 
-		w = ((u_short)s1 << 8) | (u_char)s2;
+		w = ((u_int)s1 << 8) | (u_char)s2;
 		for (i = 0; i < CNVTBLSIZ; i++)
 			if (w >= convtable[i].start
 			&& w < convtable[i].start + convtable[i].range)
@@ -506,7 +536,7 @@ static VOID NEAR j2sj(buf, s)
 char *buf;
 u_char *s;
 {
-	u_short w;
+	u_int w;
 	int i, s1, s2, j1, j2;
 
 	j1 = s[0] & 0x7f;
@@ -539,7 +569,7 @@ int max;
 
 	for (i = j = 0; s[i] && j < max - 1; i++, j++) {
 		if (isskana(s, i)) {
-			buf[j++] = 0x8e;
+			buf[j++] = (char)0x8e;
 			buf[j] = s[i];
 		}
 		else if (issjis1(s[i]) && issjis2(s[i + 1])) {
@@ -581,12 +611,12 @@ VOID readunitable(VOID_A)
 	if (!unitblpath || !*unitblpath) strcpy(path, UNICODETBL);
 	else strcatdelim2(path, unitblpath, UNICODETBL);
 
-	if ((fd = open(path, O_RDONLY | O_BINARY, 0600)) < 0) return;
+	if ((fd = open(path, O_BINARY | O_RDONLY, 0600)) < 0) return;
 	if (read(fd, buf, 2) != 2) {
 		close(fd);
 		return;
 	}
-	unitblent = (((u_short)(buf[1]) << 8) | buf[0]);
+	unitblent = (((u_int)(buf[1]) << 8) | buf[0]);
 	size = (long)unitblent * 4;
 
 	unitblbuf = (u_char *)realloc2(unitblbuf, size);
@@ -602,12 +632,13 @@ VOID discardunitable(VOID_A)
 	}
 }
 
-u_short unifysjis(wc, russ)
-u_short wc;
+u_int unifysjis(wc, russ)
+u_int wc;
 int russ;
 {
 	int i;
 
+	wc &= 0xffff;
 	for (i = ((russ) ? 0 : EXCEPTRUSS); i < RSJISTBLSIZ; i++)
 		if (wc >= rsjistable[i].start
 		&& wc < rsjistable[i].start + rsjistable[i].range)
@@ -619,15 +650,16 @@ int russ;
 	return(wc);
 }
 
-u_short cnvunicode(wc, encode)
-u_short wc;
+u_int cnvunicode(wc, encode)
+u_int wc;
 int encode;
 {
 	static int fd = -1;
 	u_char buf[4];
 	char path[MAXPATHLEN];
-	u_short r, w, min, max, ofs;
+	u_int r, w, min, max, ofs;
 
+	wc &= 0xffff;
 	if (encode < 0) {
 		if (fd >= 0) close(fd);
 		fd = -1;
@@ -672,9 +704,9 @@ int encode;
 			wc = unifysjis(wc, 0);
 			cp = unitblbuf;
 			for (ofs = 0; ofs < unitblent; ofs++) {
-				w = (((u_short)(cp[3]) << 8) | cp[2]);
+				w = (((u_int)(cp[3]) << 8) | cp[2]);
 				if (wc == w) {
-					r = (((u_short)(cp[1]) << 8) | cp[0]);
+					r = (((u_int)(cp[1]) << 8) | cp[0]);
 					break;
 				}
 				cp += 4;
@@ -687,9 +719,9 @@ int encode;
 			for (;;) {
 				if (ofs == min || ofs == max) break;
 				cp = &(unitblbuf[(ofs - 1) * 4]);
-				w = (((u_short)(cp[1]) << 8) | cp[0]);
+				w = (((u_int)(cp[1]) << 8) | cp[0]);
 				if (wc == w) {
-					r = (((u_short)(cp[3]) << 8) | cp[2]);
+					r = (((u_int)(cp[3]) << 8) | cp[2]);
 					break;
 				}
 				else if (wc < w) {
@@ -710,7 +742,7 @@ int encode;
 		if (!unitblpath || !*unitblpath) strcpy(path, UNICODETBL);
 		else strcatdelim2(path, unitblpath, UNICODETBL);
 
-		if ((fd = open(path, O_RDONLY | O_BINARY, 0600)) < 0)
+		if ((fd = open(path, O_BINARY | O_RDONLY, 0600)) < 0)
 			return(r);
 		if (!unitblent) {
 			if (read(fd, buf, 2) != 2) {
@@ -718,18 +750,18 @@ int encode;
 				fd = -1;
 				return(r);
 			}
-			unitblent = (((u_short)(buf[1]) << 8) | buf[0]);
+			unitblent = (((u_int)(buf[1]) << 8) | buf[0]);
 		}
 	}
 
 	if (encode) {
-		if (lseek(fd, (off_t)2, L_SET) < 0) return(r);
+		if (Xlseek(fd, (off_t)2, L_SET) < (off_t)0) return(r);
 		wc = unifysjis(wc, 0);
 		for (ofs = 0; ofs < unitblent; ofs++) {
 			if (read(fd, buf, 4) != 4) break;
-			w = (((u_short)(buf[3]) << 8) | buf[2]);
+			w = (((u_int)(buf[3]) << 8) | buf[2]);
 			if (wc == w) {
-				r = (((u_short)(buf[1]) << 8) | buf[0]);
+				r = (((u_int)(buf[1]) << 8) | buf[0]);
 				break;
 			}
 		}
@@ -740,11 +772,12 @@ int encode;
 		ofs = unitblent / 2 + 1;
 		for (;;) {
 			if (ofs == min || ofs == max) break;
-			if (lseek(fd, (off_t)(ofs - 1) * 4 + 2, L_SET) < 0
+			if (Xlseek(fd, (off_t)(ofs - 1) * 4 + 2, L_SET)
+			< (off_t)0
 			|| read(fd, buf, 4) != 4) break;
-			w = (((u_short)(buf[1]) << 8) | buf[0]);
+			w = (((u_int)(buf[1]) << 8) | buf[0]);
 			if (wc == w) {
-				r = (((u_short)(buf[3]) << 8) | buf[2]);
+				r = (((u_int)(buf[3]) << 8) | buf[2]);
 				break;
 			}
 			else if (wc < w) {
@@ -846,10 +879,10 @@ int max, knj, asc, io;
 	return(j);
 }
 
-static int NEAR fromjis7(buf, s, max, knj, asc, io)
+static int NEAR fromjis(buf, s, max, io)
 char *buf;
 u_char *s;
-int max, knj, asc, io;
+int max, io;
 {
 	int i, j, mode;
 
@@ -863,22 +896,52 @@ int max, knj, asc, io;
 			mode &= ~KANA;
 			break;
 		case '\033':	/* ESC */
-			if (s[i + 1] == '$' && s[i + 2] == knj) {
-				mode |= KANJI;
-				i += 2;
-				break;
+			if (s[i + 1] == '$') {
+				if (s[i + 2] == '@' || s[i + 2] == 'B') {
+					mode &= ~JKANA;
+					mode |= KANJI;
+					i += 2;
+					break;
+				}
+				else if (s[i + 2] == '('
+				&& (s[i + 3] == '@' || s[i + 3] == 'B'
+				|| s[i + 3] == 'O')) {
+					mode &= ~JKANA;
+					mode |= KANJI;
+					i += 3;
+					break;
+				}
 			}
-			else if (s[i + 1] == '(' && s[i + 2] == asc) {
-				mode &= ~KANJI;
-				i += 2;
-				break;
+			else if (s[i + 1] == '(') {
+				if (s[i + 2] == 'J' || s[i + 2] == 'B'
+				|| s[i + 2] == 'H') {
+					mode &= ~(KANJI | JKANA);
+					i += 2;
+					break;
+				}
+				else if (s[i + 2] == 'I') {
+					mode &= ~KANJI;
+					mode |= JKANA;
+					i += 2;
+					break;
+				}
 			}
+			else if (s[i + 1] == '&') {
+				if (s[i + 2] == '@' && s[i + 3] == '\033'
+				&& s[i + 4] == '$' && s[i + 5] == 'B') {
+					mode &= ~JKANA;
+					mode |= KANJI;
+					i += 5;
+					break;
+				}
+			}
+/*FALLTHRU*/
 		default:
-			if (mode & KANA) {
+			if (mode & (KANA | JKANA)) {
 				if (!isjkana(s, i)) buf[j++] = s[i];
 				else {
 # ifdef	CODEEUC
-					buf[j++] = 0x8e;
+					buf[j++] = (char)0x8e;
 # endif
 					buf[j++] = jdecnv(s[i], io) | 0x80;
 				}
@@ -902,6 +965,12 @@ int max, knj, asc, io;
 # endif
 				}
 			}
+#  ifdef	CODEEUC
+			else if (iskna(s[i])) {
+				buf[j++] = (char)0x8e;
+				buf[j++] = s[i];
+			}
+#  endif
 			else buf[j++] = s[i];
 			break;
 	}
@@ -963,59 +1032,6 @@ int max, knj, asc, io;
 	return(j);
 }
 
-static int NEAR fromjis8(buf, s, max, knj, asc, io)
-char *buf;
-u_char *s;
-int max, knj, asc, io;
-{
-	int i, j, mode;
-
-	mode = ASCII;
-	for (i = j = 0; s[i] && j < max - 1; i++)
-	switch (s[i]) {
-		case '\033':	/* ESC */
-			if (s[i + 1] == '$' && s[i + 2] == knj) {
-				mode = KANJI;
-				i += 2;
-				break;
-			}
-			else if (s[i + 1] == '(' && s[i + 2] == asc) {
-				mode = ASCII;
-				i += 2;
-				break;
-			}
-		default:
-			if (mode & KANJI) {
-				if (!isjis(s[i]) || !isjis(s[i + 1])) {
-					buf[j++] = s[i++];
-					buf[j++] = s[i];
-				}
-				else {
-#  ifdef	CODEEUC
-					buf[j++] = s[i++] | 0x80;
-					buf[j++] = jdecnv(s[i], io) | 0x80;
-#  else
-					u_char tmp[2];
-
-					tmp[0] = s[i++];
-					tmp[1] = jdecnv(s[i], io);
-					j2sj(&(buf[j]), tmp);
-					j += 2;
-#  endif
-				}
-			}
-#  ifdef	CODEEUC
-			else if (iskna(s[i])) {
-				buf[j++] = 0x8e;
-				buf[j++] = s[i];
-			}
-#  endif
-			else buf[j++] = s[i];
-			break;
-	}
-	return(j);
-}
-
 static int NEAR tojunet(buf, s, max, knj, asc, io)
 char *buf;
 u_char *s;
@@ -1073,75 +1089,12 @@ int max, knj, asc, io;
 	return(j);
 }
 
-static int NEAR fromjunet(buf, s, max, knj, asc, io)
-char *buf;
-u_char *s;
-int max, knj, asc, io;
-{
-	int i, j, mode;
-
-	mode = ASCII;
-	for (i = j = 0; s[i] && j < max - 1; i++)
-	switch (s[i]) {
-		case '\033':	/* ESC */
-			if (s[i + 1] == '$' && s[i + 2] == knj) {
-				mode = KANJI;
-				i += 2;
-				break;
-			}
-			else if (s[i + 1] == '(') {
-				if (s[i + 2] == 'I') {
-					mode = JKANA;
-					i += 2;
-					break;
-				}
-				else if (s[i + 2] == asc) {
-					mode = ASCII;
-					i += 2;
-					break;
-				}
-			}
-		default:
-			if (mode == JKANA) {
-				if (!isjkana(s, i)) buf[j++] = s[i];
-				else {
-#  ifdef	CODEEUC
-					buf[j++] = 0x8e;
-#  endif
-					buf[j++] = jdecnv(s[i], io) | 0x80;
-				}
-			}
-			else if (mode == KANJI) {
-				if (!isjis(s[i]) || !isjis(s[i + 1])) {
-					buf[j++] = s[i++];
-					buf[j++] = s[i];
-				}
-				else {
-#  ifdef	CODEEUC
-					buf[j++] = s[i++] | 0x80;
-					buf[j++] = jdecnv(s[i], io) | 0x80;
-#  else
-					u_char tmp[2];
-
-					tmp[0] = s[i];
-					tmp[1] = jdecnv(s[i], io);
-					j2sj(&(buf[j]), tmp);
-					j += 2;
-#  endif
-				}
-			}
-			else buf[j++] = s[i];
-			break;
-	}
-	return(j);
-}
-
 static int NEAR toutf8(buf, s, max)
 char *buf;
 u_char *s;
 int max;
 {
-	u_short w;
+	u_int w;
 	int i, j;
 
 	for (i = j = 0; s[i] && j < max - 3; i++, j++) {
@@ -1150,7 +1103,7 @@ int max;
 		else if (iseuc(s[i]) && iseuc(s[i + 1])) {
 			u_char tmp[2];
 
-			j2sj(tmp, &(s[i++]));
+			j2sj((char *)tmp, &(s[i++]));
 			w = (tmp[0] << 8) | tmp[1];
 		}
 #  else
@@ -1181,7 +1134,7 @@ char *buf;
 u_char *s;
 int max;
 {
-	u_short w;
+	u_int w;
 	int i, j, c1, c2;
 
 	for (i = j = 0; s[i] && j < max - 2; i++, j++) {
@@ -1250,7 +1203,7 @@ int max;
 		else if (iseuc(s[i]) && iseuc(s[i + 1])) {
 			u_char tmp[2];
 
-			j2sj(tmp, &(s[i++]));
+			j2sj((char *)tmp, &(s[i++]));
 			j += bin2hex(&(buf[j]), tmp[0]);
 			j += bin2hex(&(buf[j]), tmp[1]) - 1;
 		}
@@ -1345,7 +1298,7 @@ int max;
 		else if (iseuc(s[i]) && iseuc(s[i + 1])) {
 			u_char tmp[2];
 
-			j2sj(tmp, &(s[i++]));
+			j2sj((char *)tmp, &(s[i++]));
 			j += bin2cap(&(buf[j]), tmp[0]);
 			j += bin2cap(&(buf[j]), tmp[1]) - 1;
 		}
@@ -1440,30 +1393,17 @@ int max, in, out, *lenp, io;
 # endif
 					break;
 				case JIS7:
-					*lenp = fromjis7(buf, (u_char *)s,
-						max, 'B', 'B', io);
-					break;
 # if	defined (FD) && (FD >= 2)
 				case O_JIS7:
-					*lenp = fromjis7(buf, (u_char *)s,
-						max, '@', 'J', io);
-					break;
 				case JIS8:
-					*lenp = fromjis8(buf, (u_char *)s,
-						max, 'B', 'B', io);
-					break;
 				case O_JIS8:
-					*lenp = fromjis8(buf, (u_char *)s,
-						max, '@', 'J', io);
-					break;
 				case JUNET:
-					*lenp = fromjunet(buf, (u_char *)s,
-						max, 'B', 'B', io);
-					break;
 				case O_JUNET:
-					*lenp = fromjunet(buf, (u_char *)s,
-						max, '@', 'J', io);
+# endif	/* FD && (FD >= 2) */
+					*lenp = fromjis(buf, (u_char *)s,
+						max, io);
 					break;
+# if	defined (FD) && (FD >= 2)
 				case UTF8:
 					*lenp = fromutf8(buf, (u_char *)s,
 						max);
@@ -1602,49 +1542,45 @@ char *s;
 #ifdef	USESTDARGH
 /*VARARGS1*/
 int kanjiprintf(CONST char *fmt, ...)
-{
-	va_list args;
-	char buf[MAXPRINTBUF + 1];
-
-	va_start(args, fmt);
-	vsprintf(buf, fmt, args);
-	va_end(args);
-#else	/* !USESTDARGH */
-# ifndef	NOVSPRINTF
+#else
 /*VARARGS1*/
 int kanjiprintf(fmt, va_alist)
-char *fmt;
+CONST char *fmt;
 va_dcl
+#endif
 {
 	va_list args;
-	char buf[MAXPRINTBUF + 1];
+	char *buf;
+	int n;
 
+#ifdef	USESTDARGH
+	va_start(args, fmt);
+#else
 	va_start(args);
-	vsprintf(buf, fmt, args);
-	va_end(args);
-# else	/* NOVSPRINTF */
-int kanjiprintf(fmt, arg1, arg2, arg3, arg4, arg5, arg6)
-char *fmt;
-{
-	char buf[MAXPRINTBUF + 1];
+#endif
 
-	sprintf(buf, fmt, arg1, arg2, arg3, arg4, arg5, arg6);
-# endif	/* NOVSPRINTF */
-#endif	/* !USESTDARGH */
-	return(kanjiputs(buf));
+	n = vasprintf2(&buf, fmt, args);
+	va_end(args);
+	if (n < 0) error("malloc()");
+	n = strlen3(buf);
+	kanjiputs(buf);
+	free(buf);
+	return(n);
 }
 
 int kanjiputs2(s, len, ptr)
 char *s;
 int len, ptr;
 {
-	char *dupl;
+	char *buf;
+	int n;
 
-	if (len >= 0) dupl = malloc2(len + 1);
-	else if (ptr < 0) dupl = malloc2(strlen(s) + 1);
-	else dupl = malloc2(strlen(&(s[ptr])) + 1);
-	strncpy3(dupl, s, &len, ptr);
-	kanjiputs(dupl);
-	free(dupl);
-	return(len);
+	if (len >= 0) buf = malloc2(len * KANAWID + 1);
+	else if (ptr < 0) buf = malloc2(strlen(s) + 1);
+	else buf = malloc2(strlen(&(s[ptr])) + 1);
+	strncpy3(buf, s, &len, ptr);
+	n = strlen3(buf);
+	kanjiputs(buf);
+	free(buf);
+	return(n);
 }
