@@ -17,11 +17,13 @@
 #define	USETIMEH
 #define	USEUTIME
 #define	USEMKTIME
+#define	USESTRERROR
 #define	SENSEPERSEC	20
 # if	defined (__TURBOC__) && defined (__WIN32__)
 # define	BCC32
 # endif
 # ifdef	__GNUC__
+# define	USERESOURCEH
 # define	SIGFNCINT
 #  ifndef	DJGPP
 #  define	DJGPP	1
@@ -44,6 +46,7 @@ typedef unsigned long	u_long;
 # ifdef	__TURBOC__
 #  ifndef	BCC32
 #  define	NEAR		near
+#  define	DOS16
 #  endif
 # define	FORCEDSTDC
 # define	SIGFNCINT
@@ -51,6 +54,7 @@ typedef long	off_t;
 # endif
 # ifdef	LSI_C
 # define	NEAR		near
+# define	DOS16
 # define	FORCEDSTDC
 # endif
 # if	defined (DJGPP) || defined (BCC32)
@@ -76,6 +80,7 @@ typedef long	off_t;
 || defined (USGr4) || defined (__svr4__) || defined (__SVR4)
 # define	SVR4
 # define	OSTYPE		"SOLARIS"
+# define	EXTENDCCOPT	"-O -D_FILE_OFFSET_BITS=64"
 # include <stdio.h>
 #  if	defined (__SUNPRO_C) && defined (BSD)
 #  define	USEDIRECT
@@ -83,9 +88,12 @@ typedef long	off_t;
 #  endif
 # define	USEMANLANG
 # define	REGEXPLIB	"-lgen"
+# define	USEMKDEVH
 # define	NODNAMLEN
+# define	DNAMESIZE	1
 # define	NOTMGMTOFF
 # define	USESTATVFSH
+# define	USESTATVFS_T
 # define	USEMNTTABH
 # define	USEMKTIME
 # define	SIGFNCINT
@@ -96,6 +104,7 @@ typedef long	off_t;
 # define	USERE_COMP
 # define	USETIMELOCAL
 # define	USESYSCONF
+# define	USEWAITPID
 # define	USERESOURCEH
 # define	USESETVBUF
 # endif	/* MAXHOSTNAMELEN && !USGr4 && !__svr4__ && !__SVR4 */
@@ -104,16 +113,20 @@ typedef long	off_t;
 #if	defined (sony)
 #define	USELEAPCNT
 #define	USEMANLANG
+#define	SUPPORTSJIS
 # if	defined (USGr4) || defined (SYSTYPE_SYSV)
 # define	SVR4
 # define	OSTYPE		"NEWS_OS6"
 # define	REGEXPLIB	"-lgen"
+# define	USEMKDEVH
 # define	NODNAMLEN
+# define	DNAMESIZE	1
 # define	NOTMGMTOFF
 # define	USESTATVFSH
 # define	USEMNTTABH
 # define	USEMKTIME
 # define	USEUTIME
+# define	USETIMES
 # define	SIGFNCINT
 # define	GETTODARGS	1
 # else	/* !USGr4 && !SYSTYPE_SYSV */
@@ -135,7 +148,7 @@ typedef long	off_t;
 #  define	NOTMGMTOFF
 #  define	NOVSPRINTF
 #  define	USEGETWD
-#  define	USETIMESH
+#  define	USETIMES
 #  endif	/* !__sony && bsd43 */
 # endif	/* !USGr4 && !SYSTYPE_SYSV */
 #endif
@@ -151,29 +164,36 @@ typedef long	off_t;
 # define	TERMCAPLIB	"-lcurses"
 # define	EXTENDLIB	"-lsun"
 # endif
-#define	STRICTSTDC
+#define	USEPID_T
+#define	DECLERRLIST
+#define	USEMKDEVH
 #define	NODNAMLEN
+#define	DNAMESIZE	1
 #define	NOTMGMTOFF
 #define	USESTATFSH
 #define	STATFSARGS	4
 #define	USERE_COMP
 #define	USEMKTIME
-#define	WAITPID
+#define	USESYSCONF
+#define	USEWAITPID
 #define	USESIGPMASK
 #define	USERESOURCEH
 #define	GETPGRPVOID
 #define	USESETPGID
 #define	USESETVBUF
+#define	USESTRERROR
 #endif
 
 #if	defined (hpux) || defined (__hpux) \
 || defined (__H3050) || defined (__H3050R) || defined (__H3050RX)
 #define	SYSV
 #define	OSTYPE			"HPUX"
-#define	EXTENDCCOPT		""
+#define	EXTENDCCOPT		"-D_FILE_OFFSET_BITS=64"
 #define	USETERMINFO
 #define	TERMCAPLIB		"-lcurses"
 #define	STRICTSTDC
+#define	USEPID_T
+#define	NOSIGLIST
 #define	NOTZFILEH
 #define	USEMKNODH
 #define	NOTMGMTOFF
@@ -189,6 +209,7 @@ typedef long	off_t;
 #define	USESETPGID
 #define	USESETVBUF
 #define	SIGFNCINT
+#define	USESTRERROR
 #endif
 
 #if	defined (nec_ews) || defined (_nec_ews)
@@ -205,7 +226,9 @@ typedef long	off_t;
 #  endif
 # define	TERMCAPLIB	"-lcurses"
 # define	REGEXPLIB	"-lgen"
+# define	USEMKDEVH
 # define	NODNAMLEN
+# define	DNAMESIZE	1
 # define	NOTMGMTOFF
 # define	USESTATVFSH
 # define	USEMNTTABH
@@ -234,6 +257,7 @@ typedef long	off_t;
 #define	SYSVDIRENT
 #define	HAVETIMEZONE
 #define	USEREGCMP
+#define	USERESOURCEH
 #define	WAITKEYPAD		720
 #define	WAITMETA		720
 #endif
@@ -250,6 +274,7 @@ typedef long	off_t;
 #define	USESETENV
 #define	NOVSPRINTF
 #define	USEGETWD
+#define	USERESOURCEH
 #define	SIGARGINT
 #endif
 
@@ -263,11 +288,13 @@ typedef long	off_t;
 #define	USERE_COMP
 #define	USESETENV
 #define	USEGETWD
+#define	USERESOURCEH
 #define	SIGARGINT
 #endif
 
-#if	(defined (__alpha) || defined (alpha)) \
-&& !defined (linux) && !defined (__FreeBSD__) && !defined (__NetBSD__)
+#if	(defined (alpha) || defined (__alpha) || defined (__alpha__)) \
+&& !defined (linux) \
+&& !defined (__FreeBSD__) && !defined (__NetBSD__) && !defined (__OpenBSD__)
 #define	CODEEUC
 #define	TARUSESPACE
 #define	EXTENDLIB	"-lc_r"
@@ -280,18 +307,24 @@ typedef long	off_t;
 # define	USEMOUNTH
 # define	STATFSARGS	3
 # define	USERE_COMP
+# define	USERESOURCEH
 # else	/* !SYSTYPE_BSD */
 # define	SVR4
 # define	OSTYPE		"DECOSF1V3"
-# define	NODNAMLEN
+# define	DECLERRLIST
+# define	USELEAPCNT
+# define	USETERMIO
 # define	USESTATVFSH
 # define	USEREGCOMP
+# define	USEMKTIME
+# define	SIGFNCINT
 # endif	/* !SYSTYPE_BSD */
 #endif
 
 #if	defined (_IBMR2)
 #define	SVR4
 #define	OSTYPE			"AIX"
+#define	EXTENDCCOPT		"-O -D_LARGE_FILES -U_LARGE_FILE_API"
 #define	TERMCAPLIB		"-lcurses"
 #define	NOTZFILEH
 #define	USESELECTH
@@ -299,16 +332,20 @@ typedef long	off_t;
 #define	USETIMEH
 #define	USETERMIO
 #define	NOTMGMTOFF
-#define	USESTATFSH
-#define	STATFSARGS	4
 #define	USEMNTCTL
 #define	USERE_COMP
 # if	defined (_AIX41)
+# define	USESTATVFSH
 # define	USEMKTIME
 # define	SIGFNCINT
 # else
+# define	USESTATFSH
+# define	STATFSARGS	4
 # define	SIGARGINT
 # endif
+# if	defined (_AIX43)
+# define	NOTERMVAR
+# endif
 #endif
 
 #if	defined (ultrix)
@@ -325,6 +362,7 @@ typedef long	off_t;
 #define	USEMKTIME
 #define	USESYSCONF
 #define	USESIGPMASK
+#define	USERESOURCEH
 #endif
 
 #if	defined (_AUX_SOURCE)
@@ -347,6 +385,7 @@ typedef long	off_t;
 #define	STATFSARGS	4
 #define	USESYSDIRH
 #define	NODNAMLEN
+#define	DNAMESIZE	1
 #define	USERE_COMP
 #endif
 
@@ -356,7 +395,9 @@ typedef long	off_t;
 #define	CODEEUC
 #define	CCOUTOPT		"-o $*"
 #define	REGEXPLIB		"-lgen"
+#define	USEMKDEVH
 #define	NODNAMLEN
+#define	DNAMESIZE	1
 #define	NOTMGMTOFF
 #define	USESTATVFSH
 #define	USEMNTTABH
@@ -369,7 +410,9 @@ typedef long	off_t;
 #define	OSTYPE			"UXPDS"
 #define	CODEEUC
 #define	REGEXPLIB		"-lgen"
+#define	USEMKDEVH
 #define	NODNAMLEN
+#define	DNAMESIZE	1
 #define	NOTMGMTOFF
 #define	USESTATVFSH
 #define	USEMNTTABH
@@ -377,26 +420,6 @@ typedef long	off_t;
 #define	USEMKTIME
 #endif
 
-#if	defined (mips) && !defined (OSTYPE)
-#define	BSD43
-#define	OSTYPE			"MIPS"
-#define	CODEEUC
-# if	defined (SYSTYPE_SYSV)
-# undef	CCCOMMAND
-# define	CCCOMMAND	"/bsd43/bin/cc"
-# endif
-#define	TERMCAPLIB	"-lcurses -ltermcap"
-#define	NOERRNO
-#define	NOFILEMODE
-#define	NOUNISTDH
-#define	NOSTDLIBH
-#define	USEDIRECT
-#define	USERE_COMP
-#define	USESETENV
-#define	USEGETWD
-#define	SIGARGINT
-#endif
-
 #if	defined (NeXT)
 #define	BSD43
 #define	OSTYPE			"NEXTSTEP"
@@ -406,12 +429,19 @@ typedef long	off_t;
 #define	USEDIRECT
 #define	USERE_COMP
 #define	USEGETWD
+#define	USERESOURCEH
 #endif
 
 #if	defined (linux)
-#define	OSTYPE			"LINUX"
 #define	POSIX
+#define	OSTYPE			"LINUX"
 #define	CODEEUC
+# if	defined (alpha) || defined (__alpha) || defined (__alpha__)
+/* Linux/Alpha need not support LFS and cannot use statfs(2) with LFS */
+# define	EXTENDCCOPT		"-O"
+# else
+# define	EXTENDCCOPT		"-O -D_FILE_OFFSET_BITS=64"
+# endif
 #define	USEMANLANG
 #define	BSDINSTALL
 #define	TARUSESPACE
@@ -437,12 +467,17 @@ typedef long	off_t;
 #define	NOTZFILEH
 #define	USETIMEH
 #define	SYSVDIRENT
+#define	NODNAMLEN
 #define	HAVETIMEZONE
 #define	NOTMGMTOFF
 #define	USEREGCOMP
 #define	USESETENV
 #define	USEMKTIME
-# if	!defined (__alpha__) && !defined (__alpha) && !defined (alpha)
+# if	!defined (alpha) && !defined (__alpha) && !defined (__alpha__) \
+&& !defined (ia64) && !defined (__ia64) && !defined (__ia64__) \
+&& !defined (x86_64) && !defined (__x86_64) && !defined (__x86_64__) \
+&& !defined (s390x) && !defined (__s390x) && !defined (__s390x__) \
+&& !defined (CONFIG_ARCH_S390X)
 # define	USELLSEEK
 # endif
 #define	SIGFNCINT
@@ -597,6 +632,7 @@ typedef long	off_t;
 #define	BSD43
 #define	OSTYPE			"ORG_386BSD"
 #define	TARUSESPACE
+#define	USEPID_T
 #define	DECLSIGLIST
 #define	DECLERRLIST
 #define	USELEAPCNT
@@ -609,6 +645,27 @@ typedef long	off_t;
 #define	USERESOURCEH
 #endif
 
+#if	defined (mips) && !defined (OSTYPE)
+#define	BSD43
+#define	OSTYPE			"MIPS"
+#define	CODEEUC
+# if	defined (SYSTYPE_SYSV)
+# undef	CCCOMMAND
+# define	CCCOMMAND	"/bsd43/bin/cc"
+# endif
+#define	TERMCAPLIB	"-lcurses -ltermcap"
+#define	NOERRNO
+#define	NOFILEMODE
+#define	NOUNISTDH
+#define	NOSTDLIBH
+#define	USEDIRECT
+#define	USERE_COMP
+#define	USESETENV
+#define	USEGETWD
+#define	USERESOURCEH
+#define	SIGARGINT
+#endif
+
 /****************************************************************
  *	If your machine is not described above,	you should	*
  *	comment out below manually to apply your environment.	*
@@ -654,6 +711,7 @@ typedef long	off_t;
 
 /* #define CODEEUC	;kanji code type is EUC */
 /* #define USEMANLANG	;man(1) directory includes LANG environment value */
+/* #define SUPPORTSJIS	;cc(1) or man(1) supports Shift JIS perfectly */
 /* #define BSDINSTALL	;install(1) with option -c is valid like BSD */
 /* #define BSDINSTCMD	;command name except "install" to install like BSD */
 /* #define TARUSESPACE	;tar(1) uses space to devide file mode from UID */
@@ -674,6 +732,7 @@ typedef long	off_t;
 /* #define STRICTSTDC	;cannot allow K&R type function declaration */
 /* #define NOVOID	;cannot use type 'void' */
 /* #define NOUID_T	;uid_t, gid_t is not defined in <sys/types.h> */
+/* #define USEPID_T	;pid_t is defined in <sys/types.h> as process ID */
 /* #define DECLSIGLIST	;'sys_siglist[]' declared in <signal.h> */
 /* #define NOSIGLIST	;have not 'sys_siglist[]' in library */
 /* #define DECLERRLIST	;'sys_errlist[]' declared in <stdio.h> or <errno.h> */
@@ -690,11 +749,13 @@ typedef long	off_t;
 /* #define USESTDARGH	;use <stdarg.h> for va_list */
 /* #define USEMKDEVH	;use <sys/mkdev.h> for major()/minor() */
 /* #define USEMKNODH	;use <sys/mknod.h> for major()/minor() */
+/* #define USESGTTY	;use sgtty interface */
 /* #define USETERMIO	;use termio interface */
 /* #define USETERMIOS	;use termios interface */
 /* #define USEDIRECT	;use 'struct direct' instead of dirent */
 /* #define SYSVDIRENT	;dirent interface behaves as System V */
 /* #define NODNAMLEN	;struct dirent haven't d_namlen */
+/* #define DNAMESIZE	;size of d_name in struct dirent */
 /* #define HAVETIMEZONE	;have extern valiable 'timezone' */
 /* #define NOTMGMTOFF	;struct tm haven't tm_gmtoff */
 
@@ -705,6 +766,7 @@ typedef long	off_t;
 /* #define USEMOUNTH	;use <sys/mount.h> as header of the FS status */
 /* #define USEFSDATA	;use 'struct fs_data' as structure of hte FS status */
 
+/* #define USESTATVFS_T	;use 'statvfs_t' instead of 'struct statvfs' */
 /* #define USEFFSIZE	;'struct statfs' has 'f_fsize' as block size */
 /* #define STATFSARGS	;the number of arguments in statfs() */
 
@@ -740,15 +802,18 @@ typedef long	off_t;
 /* #define USELLSEEK	;use _llseek() for 64bits width instead of lseek() */
 /* #define USEUNAME	;use uname() instead of gethostname() */
 /* #define USEWAITPID	;use waitpid() instead of wait3() */
+/* #define USESIGACTION	;use sigaction() instead of signal() */
 /* #define USESIGPMASK	;use sigprocmask() instead of sigsetmask() */
 /* #define USERESOURCEH	;use <sys/resource.h> for resource info. */
 /* #define USEULIMITH	;use <ulimit.h> for resource info. */
-/* #define USETIMESH	;use <sys/times.h> for resource info. */
+/* #define USEGETRUSAGE	;use getrusage() for getting process time */
+/* #define USETIMES	;use times() for getting process time */
 /* #define GETPGRPVOID	;getpgrp() needs void argument */
 /* #define USESETPGID	;use setpgid() instead of setpgrp() */
 /* #define USESETVBUF	;use setvbuf() instead of setbuf() or setlinebuf() */
 /* #define SIGARGINT	;the 2nd argument function of signal() returns int */
 /* #define SIGFNCINT	;the 2nd argument function of signal() needs int */
+/* #define USESTRERROR	;use strerror() instead of sys_errlist[] */
 /* #define GETTODARGS	;the number of arguments in gettimeofday() */
 
 /* #define SENSEPERSEC	;ratio of key sense per 1 second */
@@ -771,26 +836,28 @@ typedef long	off_t;
 #endif
 
 #ifdef	POSIX
-# if	!defined (USETERMIOS) && !defined (USETERMIO)
+# if	!defined (USESGTTY) && !defined (USETERMIO) && !defined (USETERMIOS)
 # define	USETERMIOS
 # endif
+#define	USEPID_T
 #define	USESYSCONF
 #define	USEWAITPID
+#define	USESIGACTION
 #define	USESIGPMASK
 #define	USERESOURCEH
 #define	GETPGRPVOID
 #define	USESETPGID
 #define	USESETVBUF
+#define	USESTRERROR
 #endif
 
 #ifdef	SVR4
 #define	SYSV
-#define	USEMKDEVH
 #endif
 
 #ifdef	SYSV
 #define	TARUSESPACE
-# if	!defined (USETERMIOS) && !defined (USETERMIO)
+# if	!defined (USESGTTY) && !defined (USETERMIO) && !defined (USETERMIOS)
 # define	USETERMIO
 # endif
 #define	SYSVDIRENT
@@ -802,6 +869,20 @@ typedef long	off_t;
 # endif
 #endif
 
+#ifdef	USETERMIOS
+# ifdef	USETERMIO
+# undef	USETERMIO
+# endif
+#endif
+
+#if	defined (USETERMIOS) || defined (USETERMIO)
+# ifdef	USESGTTY
+# undef	USESGTTY
+# endif
+#else
+#define	USESGTTY
+#endif
+
 #if	defined (__STDC__) || defined (FORCEDSTDC)
 #define	USESTDARGH
 #endif
@@ -849,6 +930,9 @@ typedef long	off_t;
 #ifndef	EXTENDLIB
 #define	EXTENDLIB		""
 #endif
+#ifndef	DNAMESIZE
+#define	DNAMESIZE		(MAXNAMLEN + 1)
+#endif
 
 #if	defined (USESTATVFSH)
 # ifdef	USESTATFSH
@@ -933,6 +1017,11 @@ typedef long	off_t;
 #define	USEMNTENTH
 #endif
 
+#if	!defined (USEGETRUSAGE) && !defined (USETIMES) \
+&& defined (USERESOURCEH)
+#define	USEGETRUSAGE
+#endif
+
 #ifndef	GETTODARGS
 #define	GETTODARGS	2
 #endif
