@@ -1,7 +1,7 @@
 /*
  *	FD (File & Directory maintenance tool)
  *
- *	by T.Shirai <shirai@nintendo.co.jp>
+ *	by T.Shirai <shirai@unixusers.net>
  */
 
 #include <ctype.h>
@@ -12,54 +12,34 @@
 #include "kanji.h"
 #include "version.h"
 
-#ifndef	_NOORIGSHELL
-#include "system.h"
-#endif
-
 #if	MSDOS
 #include <process.h>
-#include <sys/timeb.h>
 # ifdef	DJGPP
 extern char *adjustfname __P_((char *));
 # else
 # include <dos.h>
 #  ifdef	__TURBOC__
-extern unsigned _stklen = 0x6000;
+extern unsigned _stklen = 0x5800;
 #  define	harderr_t	void
 #  else
 #  define	harderr_t	int
 #  endif
 # endif
-#else
-# ifdef	_NODOSDRIVE
-# include <sys/param.h>
-# endif
 #endif	/* !MSDOS */
 
-#ifndef	_NODOSDRIVE
-#include "dosdisk.h"
-#endif
-
-#if	defined (SIGARGINT) || defined (NOVOID)
-#define	sigarg_t	int
+#ifdef	_NOORIGSHELL
+#define	Xexit2		exit2
 #else
-#define	sigarg_t	void
+#include "system.h"
 #endif
 
-#ifdef	SIGFNCINT
-#define	sigfnc_t	int
-#else
-# ifdef	NOVOID
-# define	sigfnc_t
-# else
-# define	sigfnc_t	void
-# endif
+#if	!MSDOS && !defined (_NOORIGSHELL) && !defined (NOJOB)
+extern VOID killjob __P_((VOID_A));
 #endif
 
-#define	sigcst_t	sigarg_t (*)__P_((sigfnc_t))
-
 #ifdef	_NOORIGSHELL
 extern char **environ;
+extern char **environ2;
 #endif
 extern bindtable bindlist[];
 #ifndef	_NOARCHIVE
@@ -70,10 +50,15 @@ extern int maxarchive;
 #endif
 extern char fullpath[];
 extern char **history[];
+extern char *histfile;
 extern char *helpindex[];
+extern int sizeinfo;
 extern int subwindow;
 extern int win_x;
 extern int win_y;
+#ifndef	_NOCUSTOMIZE
+extern int custno;
+#endif
 extern char *deftmpdir;
 #ifndef	_NODOSDRIVE
 extern char *unitblpath;
@@ -120,8 +105,11 @@ static int ignore_int __P_((VOID_A));
 #ifdef	SIGQUIT
 static int ignore_quit __P_((VOID_A));
 #endif
+#ifdef	SIGCONT
+static int ignore_cont __P_((VOID_A));
+#endif
 #ifdef	SIGHUP
-static int hangup __P_((VOID_A));
+static int hanguperror __P_((VOID_A));
 #endif
 #ifdef	SIGILL
 static int illerror __P_((VOID_A));
@@ -166,10 +154,11 @@ static int wintr __P_((VOID_A));
 static int printtime __P_((VOID_A));
 #endif
 #ifdef	_NOORIGSHELL
-static int NEAR doexec __P_((char *, char *, int, char *));
+static int NEAR execruncomline __P_((char *, char *, int, char *));
 #endif
 static int NEAR initoption __P_((int, char *[], char *[]));
 static int NEAR evaloption __P_((char *[]));
+static char *NEAR searchenv __P_((char *, char *[]));
 static VOID NEAR setexecname __P_((char *));
 static VOID NEAR setexecpath __P_((char *, char *[]));
 static VOID NEAR prepareexitfd __P_((VOID_A));
@@ -183,19 +172,35 @@ int adjtty = 0;
 #endif
 int showsecond = 0;
 int hideclock = 0;
+#ifdef	SIGALRM
+int noalrm = 1;
+#endif
 u_short today[3] = {0, 0, 0};
 int fd_restricted = 0;
 #ifndef	_NOCUSTOMIZE
 char **orighelpindex = NULL;
 bindtable *origbindlist = NULL;
 # if	!MSDOS && !defined (_NOKEYMAP)
-keymaptable *origkeymaplist = NULL;
+keyseq_t *origkeymaplist = NULL;
 # endif
-#endif
+# ifndef	_NOARCHIVE
+launchtable *origlaunchlist = NULL;
+int origmaxlaunch = 0;
+archivetable *origarchivelist = NULL;
+int origmaxarchive = 0;
+# endif
+# if	!MSDOS && !defined (_NODOSDRIVE)
+devinfo *origfdtype = NULL;
+# endif
+#endif	/* !_NOCUSTOMIZE */
 int inruncom = 0;
 int fdmode = 0;
 
 static int timersec = 0;
+#ifdef	SIGWINCH
+static int winchok = 0;
+static int winched = 0;
+#endif
 
 
 #if	MSDOS && !defined (PROTECTED_MODE)
@@ -215,44 +220,60 @@ char *s;
 	int duperrno;
 
 	duperrno = errno;
-	sigvecreset();
+	sigvecset(0);
 	forcecleandir(deftmpdir, tmpfilename);
 #ifndef	_NODOSDRIVE
 	dosallclose();
 #endif
 	if (!s) s = progname;
+	stdiomode();
 	endterm();
-	inittty(1);
-	fputc('\007', stderr);
+	if (dumbterm <= 2) fputc('\007', stderr);
 	errno = duperrno;
-	perror(s);
+	if (errno) perror2(s);
+	else {
+		fputs(s, stderr);
+		fputc('\n', stderr);
+	}
+	inittty(1);
+	keyflush();
+	prepareexitfd();
 #ifndef	_NOORIGSHELL
 # if	!MSDOS && !defined (NOJOB)
 	killjob();
 # endif
-	prepareexit();
+	prepareexit(0);
 #endif
+#ifdef	DEBUG
+# if	!MSDOS
+	freeterment();
+# endif
+	if (ttyout && ttyout != stderr) fclose(ttyout);
+	else if (ttyio >= 0) close(ttyio);
+	muntrace();
+#endif
+
 	exit(2);
 }
 
 static VOID NEAR signalexit(sig)
 int sig;
 {
-	signal(sig, SIG_IGN);
+	signal2(sig, SIG_IGN);
 	forcecleandir(deftmpdir, tmpfilename);
 #ifndef	_NODOSDRIVE
 	dosallclose();
 #endif
 	endterm();
 	inittty(1);
-	signal(sig, SIG_DFL);
+	signal2(sig, SIG_DFL);
 	kill(getpid(), sig);
 }
 
 #ifdef	SIGALRM
 static int ignore_alrm(VOID_A)
 {
-	signal(SIGALRM, (sigcst_t)ignore_alrm);
+	signal2(SIGALRM, (sigcst_t)ignore_alrm);
 	return(0);
 }
 #endif
@@ -260,7 +281,7 @@ static int ignore_alrm(VOID_A)
 #ifdef	SIGWINCH
 static int ignore_winch(VOID_A)
 {
-	signal(SIGWINCH, (sigcst_t)ignore_winch);
+	signal2(SIGWINCH, (sigcst_t)ignore_winch);
 	return(0);
 }
 #endif
@@ -268,7 +289,7 @@ static int ignore_winch(VOID_A)
 #ifdef	SIGINT
 static int ignore_int(VOID_A)
 {
-	signal(SIGINT, (sigcst_t)ignore_int);
+	signal2(SIGINT, (sigcst_t)ignore_int);
 	return(0);
 }
 #endif
@@ -276,13 +297,24 @@ static int ignore_int(VOID_A)
 #ifdef	SIGQUIT
 static int ignore_quit(VOID_A)
 {
-	signal(SIGQUIT, (sigcst_t)ignore_quit);
+	signal2(SIGQUIT, (sigcst_t)ignore_quit);
+	return(0);
+}
+#endif
+
+#ifdef	SIGCONT
+static int ignore_cont(VOID_A)
+{
+	signal2(SIGCONT, (sigcst_t)ignore_cont);
+# if	!MSDOS
+	suspended = 1;
+# endif
 	return(0);
 }
 #endif
 
 #ifdef	SIGHUP
-static int hangup(VOID_A)
+static int hanguperror(VOID_A)
 {
 	signalexit(SIGHUP);
 	return(0);
@@ -385,17 +417,77 @@ static int xsizerror(VOID_A)
 }
 #endif
 
+VOID checkscreen(xmax, ymax)
+int xmax, ymax;
+{
+#ifdef	SIGWINCH
+	int dupwinchok;
+#endif
+	char *cp;
+	int i, row;
+
+#ifdef	SIGWINCH
+	dupwinchok = winchok;
+	winchok = 0;
+#endif
+	for (i = 0;; i++) {
+		if (!(cp = getwsize(xmax, ymax))) {
+#ifndef	_NOTREE
+			if (treepath) row = WFILEMINTREE;
+			else
+#endif
+#ifndef	_NOCUSTOMIZE
+			if (custno >= 0) row = WFILEMINCUSTOM;
+			else
+#endif
+			row = WFILEMIN;
+
+			if (FILEPERROW >= row) break;
+			cp = NOROW_K;
+		}
+		if (!i) {
+			putterm(t_clear);
+			locate(0, 0);
+			keyflush();
+		}
+
+		if (i & 1) cputs2(SCRSZ_K);
+		else {
+			putterm(t_standout);
+			cputs2(cp);
+			putterm(end_standout);
+		}
+		cputs2("\r\n");
+		putterm(t_bell);
+		tflush();
+		keyflush();
+		if (kbhit2(1000000L) && getkey2(0) == K_ESC) {
+			errno = 0;
+			error(INTR_K);
+		}
+	}
+#ifdef	SIGWINCH
+	winchok = dupwinchok;
+#endif
+}
+
 #ifdef	SIGWINCH
 static int wintr(VOID_A)
 {
-	int duperrno;
+	int duperrno, dupusegetcursor;
 
 	duperrno = errno;
-	signal(SIGWINCH, SIG_IGN);
-	getwsize(80, WHEADERMAX + WFOOTER + WFILEMIN);
-	rewritefile(1);
-	if (subwindow) ungetch2(CTRL('L'));
-	signal(SIGWINCH, (sigcst_t)wintr);
+	signal2(SIGWINCH, SIG_IGN);
+	if (!winchok) winched++;
+	else {
+		dupusegetcursor = usegetcursor;
+		usegetcursor = 0;
+		checkscreen(WCOLUMNMIN, WHEADERMAX + WFOOTER + WFILEMIN);
+		usegetcursor = dupusegetcursor;
+		rewritefile(1);
+		if (subwindow) ungetch2(K_CTRL('L'));
+	}
+	signal2(SIGWINCH, (sigcst_t)wintr);
 	errno = duperrno;
 	return(0);
 }
@@ -406,25 +498,13 @@ static int printtime(VOID_A)
 {
 	static time_t now;
 	struct tm *tm;
-#if	MSDOS
-	struct timeb buffer;
-#else
-	struct timeval t_val;
-	struct timezone tz;
-#endif
-	int duperrno;
+	int x, duperrno;
 
 	duperrno = errno;
-	signal(SIGALRM, SIG_IGN);
+	signal2(SIGALRM, SIG_IGN);
 	if (timersec) now++;
 	else {
-#if	MSDOS
-		ftime(&buffer);
-		now = (time_t)(buffer.time);
-#else
-		gettimeofday2(&t_val, &tz);
-		now = (time_t)(t_val.tv_sec);
-#endif
+		now = time2();
 		timersec = CLOCKUPDATE;
 	}
 	if (showsecond || timersec == CLOCKUPDATE) {
@@ -443,7 +523,9 @@ static int printtime(VOID_A)
 		today[1] = tm -> tm_mon;
 		today[2] = tm -> tm_mday;
 		if (!hideclock) {
-			locate(n_column - 16 - ((showsecond) ? 3 : 0), LTITLE);
+			x = n_column - 15 - ((showsecond) ? 3 : 0);
+			if (!isleftshift()) x--;
+			locate(x, L_TITLE);
 			putterm(t_standout);
 			cprintf2("%02d-%02d-%02d %02d:%02d",
 				tm -> tm_year % 100,
@@ -456,54 +538,90 @@ static int printtime(VOID_A)
 		}
 	}
 	timersec--;
-	signal(SIGALRM, (sigcst_t)printtime);
+#ifdef	SIGWINCH
+	if (!winchok) winchok++;
+	else if (winched) {
+		signal2(SIGWINCH, SIG_IGN);
+		winched = 0;
+		checkscreen(WCOLUMNMIN, WHEADERMAX + WFOOTER + WFILEMIN);
+		rewritefile(1);
+		if (subwindow) ungetch2(K_CTRL('L'));
+		signal2(SIGWINCH, (sigcst_t)wintr);
+	}
+#endif
+	signal2(SIGALRM, (sigcst_t)printtime);
 	errno = duperrno;
 	return(0);
 }
 #endif
 
-VOID sigvecset(VOID_A)
+int sigvecset(set)
+int set;
 {
-#ifdef	SIGALRM
-	signal(SIGALRM, (sigcst_t)printtime);
-#endif
-#ifdef	SIGTSTP
-	signal(SIGTSTP, SIG_IGN);
-#endif
-#ifdef	SIGWINCH
-	signal(SIGWINCH, (sigcst_t)wintr);
-#endif
-#ifndef	_NODOSDRIVE
-	doswaitfunc = waitmes;
-	dosintrfunc = intrkey;
-#endif
-}
+	static int status = -1;
+	int old;
 
-VOID sigvecreset(VOID_A)
-{
+	old = status;
+	if (set == old) /*EMPTY*/;
+	else if (set) {
 #ifdef	SIGALRM
-	signal(SIGALRM, (sigcst_t)ignore_alrm);
+		signal2(SIGALRM, (sigcst_t)printtime);
+		noalrm = 0;
 #endif
 #ifdef	SIGTSTP
-	signal(SIGTSTP, SIG_DFL);
+		signal2(SIGTSTP, SIG_IGN);
 #endif
 #ifdef	SIGWINCH
-	signal(SIGWINCH, (sigcst_t)ignore_winch);
+		winchok = winched = 0;
+		signal2(SIGWINCH, (sigcst_t)wintr);
 #endif
 #ifndef	_NODOSDRIVE
-	doswaitfunc = NULL;
-	dosintrfunc = NULL;
+		doswaitfunc = waitmes;
+		dosintrfunc = intrkey;
 #endif
+		status = 1;
+	}
+	else {
+#ifdef	SIGALRM
+		signal2(SIGALRM, (sigcst_t)ignore_alrm);
+		noalrm = 1;
+#endif
+#ifdef	SIGTSTP
+		signal2(SIGTSTP, SIG_DFL);
+#endif
+#ifdef	SIGWINCH
+		signal2(SIGWINCH, (sigcst_t)ignore_winch);
+#endif
+#ifndef	_NODOSDRIVE
+		doswaitfunc = NULL;
+		dosintrfunc = NULL;
+#endif
+		status = 0;
+	}
+
+	return(old);
 }
 
 VOID title(VOID_A)
 {
 	char *cp, *eol;
-	int i;
+	int i, len;
 
-	locate(0, LTITLE);
+	locate(0, L_TITLE);
 	putterm(t_standout);
-	cputs2("  FD(File & Directory tool) Ver.");
+	len = 0;
+	if (!isleftshift()) {
+		putch2(' ');
+		len++;
+	}
+	cputs2(" FD");
+	len += 3;
+	if (!ishardomit()) {
+		cputs2("(File & Directory tool)");
+		len += 23;
+	}
+	cputs2(" Ver.");
+	len += 5;
 	cp = strchr(version, ' ');
 	while (*(++cp) == ' ');
 	if (!(eol = strchr(cp, ' '))) eol = cp + strlen(cp);
@@ -513,9 +631,9 @@ VOID title(VOID_A)
 		putch2('#');
 		i++;
 	}
-	cp = " (c)1995-2002 T.Shirai  ";
+	cp = (iswellomit()) ? "" : " (c)1995-2003 T.Shirai  ";
 	cputs2(cp);
-	i = n_column - 32 - strlen(cp) - i;
+	i = n_column - len - strlen(cp) - i;
 	while (i-- > 0) putch2(' ');
 	putterm(end_standout);
 	timersec = 0;
@@ -524,8 +642,28 @@ VOID title(VOID_A)
 #endif
 }
 
+#ifndef	_NOCUSTOMIZE
+VOID saveorigenviron(VOID_A)
+{
+	orighelpindex = copystrarray(NULL, helpindex, NULL, 10);
+	origbindlist = copybind(NULL, bindlist);
+# if	!MSDOS && !defined (_NOKEYMAP)
+	origkeymaplist = copykeyseq(NULL);
+# endif
+# ifndef	_NOARCHIVE
+	origlaunchlist = copylaunch(NULL, launchlist,
+		&origmaxlaunch, maxlaunch);
+	origarchivelist = copyarch(NULL, archivelist,
+		&origmaxarchive, maxarchive);
+# endif
+# if	!MSDOS && !defined (_NODOSDRIVE)
+	origfdtype = copydosdrive(NULL, fdtype);
+# endif
+}
+#endif	/* !_NOCUSTOMIZE */
+
 #ifdef	_NOORIGSHELL
-static int NEAR doexec(command, file, n, line)
+static int NEAR execruncomline(command, file, n, line)
 char *command, *file;
 int n;
 char *line;
@@ -533,7 +671,7 @@ char *line;
 	char *cp;
 	int i;
 
-	if (!(*(cp = skipspace(command)))) i = 0;
+	if (!*(cp = skipspace(command))) i = 0;
 	else i = execpseudoshell(cp, 0, 1);
 	if (i) {
 		putterm(l_clear);
@@ -555,14 +693,14 @@ int exist;
 #ifdef	_NOORIGSHELL
 	FILE *fp;
 	char *cp, *tmp, *fold, *line;
-	int n, cont;
+	int cont;
 #endif
-	int er;
+	int n, er;
 
 #ifdef	_NOORIGSHELL
 # if	!MSDOS
 	tmp = NULL;
-	if (!exist && (tmp = (char *)getenv("TERM"))) {
+	if (!exist && (tmp = getconstvar("TERM"))) {
 		struct stat st;
 
 		cp = malloc2(strlen(file) + strlen(tmp) + 1 + 1);
@@ -598,7 +736,7 @@ int exist;
 
 		cp = line + strlen(line);
 		for (cp--; cp >= line && (*cp == ' ' || *cp == '\t'); cp--);
-		*(cp + 1) = '\0';
+		cp[1] = '\0';
 
 		cont = 0;
 		if (cp >= line && *cp == META
@@ -619,20 +757,18 @@ int exist;
 			continue;
 		}
 
-		if (doexec(fold, tmp, n, line) < 0) er++;
+		if (execruncomline(fold, tmp, n, line) < 0) er++;
 		if (fold != line) free(line);
 		fold = NULL;
 	}
 
-	if (fold && doexec(fold, tmp, n, line) < 0) er++;
+	if (fold && execruncomline(fold, tmp, n, line) < 0) er++;
 	fclose(fp);
 	free(tmp);
 #else	/* !_NOORIGSHELL */
-	file = evalpath(strdup2(file), 1);
-	stdiomode();
+	n = stdiomode();
 	er = execruncom(file, 1);
-	ttyiomode();
-	free(file);
+	ttyiomode(n);
 #endif	/* !_NOORIGSHELL */
 
 	return(er ? -1 : 0);
@@ -656,9 +792,8 @@ char *argv[], *envp[];
 		else if (argv[i][0] != '-' || !argv[i][1]
 		|| (argv[i][1] == '-' && !argv[i][2])) break;
 		else for (cp = &(argv[i][1]); *cp; cp++) {
-			if (*cp != '_' && !isalpha(*cp)
-			&& (cp == &(argv[i][1]) || *cp < '0' || *cp > '9'))
-				break;
+			if (!isidentchar(*cp)
+			&& (cp == &(argv[i][1]) || !isdigit(*cp))) break;
 		}
 		if (cp <= &(argv[i][1]) || *cp != '=') {
 			optv[optc++] = argv[i];
@@ -669,11 +804,8 @@ char *argv[], *envp[];
 		}
 	}
 
-#ifdef	_NOORIGSHELL
-	inittty(0);
-	getterment();
-#else
-	if (initshell(optc, optv, envp) < 0) exit2(RET_FAIL);
+#ifndef	_NOORIGSHELL
+	if (initshell(optc, optv) < 0) Xexit2(RET_FAIL);
 #endif	/* !_NOORIGSHELL */
 	free(optv);
 	return(argc);
@@ -697,24 +829,30 @@ char *argv[];
 	return(i);
 }
 
+static char *NEAR searchenv(s, envp)
+char *s, *envp[];
+{
+	int i, len;
+
+	len = strlen(s);
+	for (i = 0; envp[i]; i++)
+		if (!strnpathcmp(envp[i], s, len) && envp[i][len] == '=')
+			return(&(envp[i][++len]));
+	return(NULL);
+}
+
 static VOID NEAR setexecname(argv)
 char *argv;
 {
-	char buf[MAXNAMLEN + 1];
+#if	MSDOS
+	char *cp;
+#endif
 
 	if ((progname = strrdelim(argv, 1))) progname++;
 	else progname = argv;
 #if	MSDOS
-	{
-		char *cp;
-
-		if ((cp = strchr(progname, '.'))) *cp = '\0';
-	}
-	sprintf(buf, "FD%ld", (long)getpid());
-#else
-	sprintf(buf, "%s%ld", progname, (long)getpid());
+	if ((cp = strchr(progname, '.'))) *cp = '\0';
 #endif
-	tmpfilename = strdup2(buf);
 }
 
 static VOID NEAR setexecpath(argv, envp)
@@ -723,21 +861,22 @@ char *argv, *envp[];
 	char *cp, buf[MAXPATHLEN];
 
 	origpath = getwd2();
+	if ((cp = searchenv("PWD", envp))) {
+		*fullpath = '\0';
+		realpath2(cp, fullpath, 0);
+		realpath2(fullpath, buf, 1);
+		if (!strpathcmp(origpath, buf)) {
+			free(origpath);
+			origpath = strdup2(fullpath);
+		}
+	}
 	strcpy(fullpath, origpath);
 
 #if	MSDOS
 	cp = argv;
 #else
 	if (strdelim(argv, 0)) cp = argv;
-	else {
-		int i;
-
-		for (i = 0; envp[i]; i++)
-			if (!strncmp(envp[i], "PATH=", sizeof("PATH=") - 1))
-				break;
-		if (!envp[i]) cp = NULL;
-		else cp = searchpath(argv, envp[i] + sizeof("PATH=") - 1);
-	}
+	else if ((cp = searchenv("PATH", envp))) cp = searchpath(argv, cp);
 	if (!cp) progpath = strdup2(origpath);
 	else
 #endif
@@ -751,7 +890,14 @@ char *argv, *envp[];
 
 static VOID NEAR prepareexitfd(VOID_A)
 {
-	if (_chdir2(origpath) < 0) error(origpath);
+	char cwd[MAXPATHLEN];
+
+	cwd[0] = '\0';
+	if (origpath && _chdir2(origpath) < 0) {
+		perror2(origpath);
+		if (!Xgetwd(cwd)) *cwd = '\0';
+		rawchdir(_SS_);
+	}
 	free(origpath);
 #ifndef	_NODOSDRIVE
 	dosallclose();
@@ -759,17 +905,37 @@ static VOID NEAR prepareexitfd(VOID_A)
 	free(progpath);
 
 #ifdef	DEBUG
-	free(tmpfilename);
+	if (tmpfilename) free(tmpfilename);
 # ifdef	_NOORIGSHELL
-	freeenv();
+	freevar(environ);
+	freevar(environ2);
 # endif
 # ifndef	_NOCUSTOMIZE
-	freevar(orighelpindex);
+	if (orighelpindex) {
+		freestrarray(orighelpindex, 10);
+		free(orighelpindex);
+	}
 	free(origbindlist);
 #  if	!MSDOS && !defined (_NOKEYMAP)
-	freekeymap(origkeymaplist);
+	freekeyseq(origkeymaplist);
 #  endif
-# endif
+#  ifndef	_NOARCHIVE
+	if (origlaunchlist) {
+		freelaunch(origlaunchlist, origmaxlaunch);
+		free(origlaunchlist);
+	}
+	if (origarchivelist) {
+		freearch(origarchivelist, origmaxarchive);
+		free(origarchivelist);
+	}
+#  endif
+#  if	!MSDOS && !defined (_NODOSDRIVE)
+	if (origfdtype) {
+		freedosdrive(origfdtype);
+		free(origfdtype);
+	}
+#  endif
+# endif	/* !_NOCUSTOMIZE */
 	freeenvpath();
 	freehistory(0);
 	freehistory(1);
@@ -777,19 +943,22 @@ static VOID NEAR prepareexitfd(VOID_A)
 # if	!MSDOS
 	freeidlist();
 # endif
+	chdir2(NULL);
 # if	!defined (_NOUSEHASH) && defined (_NOORIGSHELL)
 	searchhash(NULL, NULL, NULL);
 # endif
+# ifndef	_NOROCKRIDGE
+	detranspath(NULL, NULL);
+# endif
 #endif	/* DEBUG */
+	if (*cwd) rawchdir(cwd);
 }
 
 int main(argc, argv, envp)
 int argc;
 char *argv[], *envp[];
 {
-#ifndef	_NOORIGSHELL
 	char *cp;
-#endif
 	int i;
 
 #ifdef	DEBUG
@@ -806,51 +975,54 @@ char *argv[], *envp[];
 #endif
 
 #ifdef	SIGHUP
-	signal(SIGHUP, (sigcst_t)hangup);
+	signal2(SIGHUP, (sigcst_t)hanguperror);
 #endif
 #ifdef	SIGINT
-	signal(SIGINT, (sigcst_t)ignore_int);
+	signal2(SIGINT, (sigcst_t)ignore_int);
 #endif
 #ifdef	SIGQUIT
-	signal(SIGQUIT, (sigcst_t)ignore_quit);
+	signal2(SIGQUIT, (sigcst_t)ignore_quit);
+#endif
+#ifdef	SIGCONT
+	signal2(SIGCONT, (sigcst_t)ignore_cont);
 #endif
 #ifdef	SIGILL
-	signal(SIGILL, (sigcst_t)illerror);
+	signal2(SIGILL, (sigcst_t)illerror);
 #endif
 #ifdef	SIGTRAP
-	signal(SIGTRAP, (sigcst_t)traperror);
+	signal2(SIGTRAP, (sigcst_t)traperror);
 #endif
 #ifdef	SIGIOT
-	signal(SIGIOT, (sigcst_t)ioerror);
+	signal2(SIGIOT, (sigcst_t)ioerror);
 #endif
 #ifdef	SIGEMT
-	signal(SIGEMT, (sigcst_t)emuerror);
+	signal2(SIGEMT, (sigcst_t)emuerror);
 #endif
 #ifdef	SIGFPE
-	signal(SIGFPE, (sigcst_t)floaterror);
+	signal2(SIGFPE, (sigcst_t)floaterror);
 #endif
 #ifdef	SIGBUS
-	signal(SIGBUS, (sigcst_t)buserror);
+	signal2(SIGBUS, (sigcst_t)buserror);
 #endif
 #ifdef	SIGSEGV
-	signal(SIGSEGV, (sigcst_t)segerror);
+	signal2(SIGSEGV, (sigcst_t)segerror);
 #endif
 #ifdef	SIGSYS
-	signal(SIGSYS, (sigcst_t)syserror);
+	signal2(SIGSYS, (sigcst_t)syserror);
 #endif
 #ifdef	SIGPIPE
-	signal(SIGPIPE, (sigcst_t)pipeerror);
+	signal2(SIGPIPE, (sigcst_t)pipeerror);
 #endif
 #ifdef	SIGTERM
-	signal(SIGTERM, (sigcst_t)terminate);
+	signal2(SIGTERM, (sigcst_t)terminate);
 #endif
 #ifdef	SIGXCPU
-	signal(SIGXCPU, (sigcst_t)xcpuerror);
+	signal2(SIGXCPU, (sigcst_t)xcpuerror);
 #endif
 #ifdef	SIGXFSZ
-	signal(SIGXFSZ, (sigcst_t)xsizerror);
+	signal2(SIGXFSZ, (sigcst_t)xsizerror);
 #endif
-	sigvecreset();
+	sigvecset(0);
 
 #ifndef	_NOARCHIVE
 	for (maxlaunch = 0; maxlaunch < MAXLAUNCHTABLE; maxlaunch++) {
@@ -861,7 +1033,11 @@ char *argv[], *envp[];
 			strdup2(launchlist[maxlaunch].comm);
 # if	FD >= 2
 		launchlist[maxlaunch].format =
-			strdup2(launchlist[maxlaunch].format);
+			duplvar(launchlist[maxlaunch].format, -1);
+		launchlist[maxlaunch].lignore =
+			duplvar(launchlist[maxlaunch].lignore, -1);
+		launchlist[maxlaunch].lerror =
+			duplvar(launchlist[maxlaunch].lerror, -1);
 # endif
 	}
 	for (maxarchive = 0; maxarchive < MAXARCHIVETABLE; maxarchive++) {
@@ -892,18 +1068,21 @@ char *argv[], *envp[];
 	environ[i] = NULL;
 #endif
 
+#ifdef	_NOORIGSHELL
+	inittty(0);
+	getterment(NULL);
+#else
+	setshellvar(envp);
+	prepareterm();
+#endif
 	setexecpath(argv[0], envp);
 #ifndef	_NODOSDRIVE
+# ifdef	DATADIR
+	unitblpath = DATADIR;
+# else
 	unitblpath = progpath;
-#endif
-
-#ifndef	_NOCUSTOMIZE
-	orighelpindex = copystrarray(NULL, helpindex, NULL, 10);
-	origbindlist = copybind(NULL, bindlist);
-# if	!MSDOS && !defined (_NOKEYMAP)
-	origkeymaplist = copykeymap(NULL);
 # endif
-#endif	/* !_NOCUSTOMIZE */
+#endif
 
 #ifndef	_NOORIGSHELL
 	cp = progname;
@@ -913,26 +1092,36 @@ char *argv[], *envp[];
 # else
 	if (*cp == 'r') cp++;
 # endif
-	if (!strpathcmp(cp, FDSHELL)) {
-		if (!Xgetwd(fullpath)) exit2(1);
+	if (!strpathcmp(cp, FDSHELL) || !strpathcmp(cp, "su")) {
 		i = main_shell(argc, argv, envp);
 		prepareexitfd();
-		exit2(i);
+		Xexit2(i);
+	}
+	if (dumbterm > 1) {
+		errno = 0;
+		error(NTERM_K);
 	}
 #endif
 
 	fdmode = 1;
 	argc = initoption(argc, argv, envp);
-	ttyiomode();
+	ttyiomode(0);
 	initterm();
-	getwsize(80, WHEADERMAX + WFOOTER + WFILEMIN);
+	if ((cp = getwsize(WCOLUMNMIN, WHEADERMAX + WFOOTER + WFILEMIN))) {
+		errno = 0;
+		error(cp);
+	}
+#ifndef	_NOCUSTOMIZE
+	saveorigenviron();
+#endif
 
 	locate(0, n_line - 1);
 	inruncom = 1;
+/*NOTDEFINED*/
 	i = loadruncom(DEFRUNCOM, 0);
 	i += loadruncom(RUNCOMFILE, 0);
 	inruncom = 0;
-	sigvecset();
+	sigvecset(1);
 	if (i < 0) {
 		hideclock = 1;
 		warning(0, HITKY_K);
@@ -946,16 +1135,19 @@ char *argv[], *envp[];
 	if (adjtty) {
 		stdiomode();
 		inittty(0);
-		ttyiomode();
+		ttyiomode(0);
 	}
 #endif	/* !MSDOS */
 
-	loadhistory(0, HISTORYFILE);
+	loadhistory(0, histfile);
 	entryhist(1, origpath, 1);
 	putterms(t_clear);
 
+#ifdef	SIGWINCH
+	winchok = 1;
+#endif
 	main_fd(argv[i]);
-	sigvecreset();
+	sigvecset(0);
 	prepareexitfd();
 
 	stdiomode();
@@ -963,8 +1155,7 @@ char *argv[], *envp[];
 # if	!MSDOS && !defined (NOJOB)
 	killjob();
 # endif
-	prepareexit();
 #endif
-	exit2(0);
+	Xexit2(0);
 	return(0);
 }
