@@ -4,7 +4,7 @@
  *	Tree Mode Module
  */
 
-#include <signal.h>
+#include <ctype.h>
 #include "fd.h"
 #include "term.h"
 #include "func.h"
@@ -12,13 +12,8 @@
 
 #ifndef	_NOTREE
 
-#if	MSDOS
-#include <ctype.h>
-# ifndef	_NODOSDRIVE
+#if	MSDOS && !defined (_NODOSDRIVE)
 extern int preparedrv __P_((int));
-# endif
-#else
-#include <sys/param.h>
 #endif
 
 
@@ -101,7 +96,7 @@ int disp;
 	if (!(dirp = Xopendir(dir))) return(0);
 	while ((dp = Xreaddir(dirp))) {
 		if (isdotdir(dp -> d_name)) continue;
-		strcpy(path + len, dp -> d_name);
+		strcpy(&(path[len]), dp -> d_name);
 		if (limit-- <= 0
 		|| (stat2(path, &st) >= 0
 		&& (st.st_mode & S_IFMT) == S_IFDIR)) {
@@ -138,7 +133,10 @@ int level, *maxp;
 	DIR *dirp;
 	struct dirent *dp;
 	struct stat st;
-	char *cp, *dir, *subdir;
+	char *cp, *dir, *subdir, cwd[MAXPATHLEN];
+#if	!MSDOS && !defined (_NODOSDRIVE)
+	char tmp[MAXPATHLEN];
+#endif
 	int i, len;
 
 	if ((level + 1) * DIRFIELD + 2 > TREEFIELD) {
@@ -149,6 +147,11 @@ int level, *maxp;
 		*maxp = -1;
 		return(NULL);
 	}
+	if (!Xgetwd(cwd)) {
+		*maxp = 0;
+		return(NULL);
+	}
+
 #if	MSDOS
 	if (_dospath(path)) path += 2;
 #endif
@@ -159,16 +162,14 @@ int level, *maxp;
 	}
 #if	!MSDOS && !defined (_NODOSDRIVE)
 	else if (_dospath(path)) {
-		dir = malloc2(3 + 1);
-		strncpy2(dir, path, 3);
+		dir = strndup2(path, 3);
 		subdir = path + 3;
 		if (!*subdir) subdir = NULL;
 	}
 #endif
 	else {
 		len = (cp = strdelim(path, 0)) ? cp - path : strlen(path);
-		dir = malloc2(len + 1);
-		strncpy2(dir, path, len);
+		dir = strndup2(path, len);
 		subdir = (cp) ? cp + 1 : NULL;
 	}
 
@@ -183,7 +184,7 @@ int level, *maxp;
 	i = 0;
 	while ((dp = Xreaddir(dirp))) {
 		if (isdotdir(dp -> d_name)
-		|| Xstat(dp -> d_name, &st) < 0
+		|| Xstat(nodospath(tmp, dp -> d_name), &st) < 0
 		|| (st.st_mode & S_IFMT) != S_IFDIR) continue;
 		list = b_realloc(list, *maxp, treelist);
 		if (!subdir) {
@@ -240,8 +241,11 @@ int level, *maxp;
 		if (list[i].sub)
 			list[i].sub = maketree(subdir, NULL, &(list[i]),
 				level + 1, &(list[i].max));
-		else if (list[i].max >= 0 && evaldir(list[i].name, 0))
-			list[i].max = -1;
+		else {
+			if (list[i].max >= 0
+			&& evaldir(nodospath(tmp, list[i].name), 0))
+				list[i].max = -1;
+		}
 	}
 	if (sorttree && sorton) qsort(list, *maxp, sizeof(treelist), cmptree);
 
@@ -250,14 +254,14 @@ int level, *maxp;
 	if (*path == _SC_) {
 #else	/* !MSDOS */
 	if (*path == _SC_
-#ifndef	_NODOSDRIVE
+# ifndef	_NODOSDRIVE
 	|| _dospath(path)
-#endif
+# endif
 	) {
 #endif	/* !MSDOS */
 		if (_chdir2(fullpath) < 0) error(fullpath);
 	}
-	else if (strcmp(path, ".") && _chdir2("..") < 0) error("..");
+	else if (strcmp(path, ".") && _chdir2(cwd) < 0) error("..");
 	return(list);
 }
 
@@ -309,8 +313,10 @@ static VOID NEAR showtree(VOID_A)
 		locate(0, i);
 		putterm(l_clear);
 	}
-	for (i = LFILETOP + 1; i < LFILEBOTTOM; i++)
-		sprintf(bufptr(i), "%-*.*s", TREEFIELD, TREEFIELD, " ");
+	for (i = LFILETOP + 1; i < LFILEBOTTOM; i++) {
+		memset(bufptr(i), ' ', TREEFIELD);
+		bufptr(i)[TREEFIELD] = '\0';
+	}
 	_showtree(tr_root -> sub, 1, 0, tr_top);
 	for (i = LFILETOP + 1; i < LFILEBOTTOM; i++) {
 		locate(1, i);
@@ -327,13 +333,13 @@ static VOID NEAR treebar(VOID_A)
 	locate(1, LFILETOP);
 	cputs2("Tree=");
 	kanjiputs2(treepath, n_column - 6, 0);
-	locate(0, LMESLINE);
+	locate(0, L_MESLINE);
 	putterm(l_clear);
 }
 
 VOID rewritetree(VOID_A)
 {
-	tr_scr = realloc2(tr_scr, (FILEPERLOW - 1) * (TREEFIELD + 1));
+	tr_scr = realloc2(tr_scr, (FILEPERROW - 1) * (TREEFIELD + 1));
 	searchtree();
 	showtree();
 	treebar();
@@ -369,18 +375,18 @@ int max, nest;
 				for (j = strlen(treepath); j >= 0; j--)
 					treepath[j + len] = treepath[j];
 				memcpy(treepath, list[i].name, len - 1);
-#else
+#else	/* !MSDOS */
 				if (*list[i].name != _SC_
-#ifndef	_NODOSDRIVE
+# ifndef	_NODOSDRIVE
 				&& !_dospath(list[i].name)
-#endif
+# endif
 				) len++;
 				for (j = strlen(treepath); j >= 0; j--)
 					treepath[j + len] = treepath[j];
 				if (*list[i].name != _SC_)
 					memcpy(treepath, list[i].name,
 						len - 1);
-#endif
+#endif	/* !MSDOS */
 				treepath[len - 1] = _SC_;
 			}
 		}
@@ -546,13 +552,9 @@ static int NEAR _tree_input(VOID_A)
 	int ch, tmp, half;
 
 	keyflush();
-#ifdef	_NOEDITMODE
-	ch = Xgetkey(SIGALRM);
-#else
-	Xgetkey(-1);
-	ch = Xgetkey(SIGALRM);
-	Xgetkey(-1);
-#endif
+	Xgetkey(-1, 0);
+	ch = Xgetkey(1, 0);
+	Xgetkey(-1, 0);
 
 	old = tr_cur;
 	switch (ch) {
@@ -563,7 +565,7 @@ static int NEAR _tree_input(VOID_A)
 			treedown();
 			break;
 		case K_PPAGE:
-			half = (FILEPERLOW - 1) / 2;
+			half = (FILEPERROW - 1) / 2;
 			tmp = LFILETOP + half + 1 - tr_line;
 			if (tr_top + half > LFILETOP + 1)
 				half = LFILETOP - tr_top + 1;
@@ -573,7 +575,7 @@ static int NEAR _tree_input(VOID_A)
 			searchtree();
 			break;
 		case K_NPAGE:
-			half = (FILEPERLOW - 1) / 2;
+			half = (FILEPERROW - 1) / 2;
 			tmp = half + (LFILETOP + half + 1) - tr_line;
 			while (tmp-- > 0) if (treedown() < 0) break;
 			tmp = tr_line - (LFILETOP + half + 1);
@@ -640,6 +642,7 @@ static int NEAR _tree_input(VOID_A)
 				redraw = 1;
 				break;
 			}
+/*FALLTHRU*/
 		case K_BS:
 			do {
 				if (treeup() < 0) break;
@@ -656,7 +659,7 @@ static int NEAR _tree_input(VOID_A)
 			ch = '\0';
 			free(cwd);
 			break;
-		case CTRL('L'):
+		case K_CTRL('L'):
 			rewritefile(1);
 			break;
 		case K_ESC:
@@ -681,6 +684,9 @@ static char *NEAR _tree(VOID_A)
 {
 #if	!MSDOS
 	struct stat st;
+# ifndef	_NODOSDRIVE
+	char tmp[MAXPATHLEN];
+# endif
 #endif
 	char *cp, *cwd, path[MAXPATHLEN];
 	int ch, oy, otop;
@@ -701,18 +707,19 @@ static char *NEAR _tree(VOID_A)
 	strcpy(path, fullpath);
 	strncpy2(tr_cur[0].name, path, 3);
 #else	/* !MSDOS */
-# ifndef	_NODOSDRIVE
-	if (dospath("", path)) {
-		tr_cur[0].name = malloc2(3 + 1);
-		strncpy2(tr_cur[0].name, path, 3);
-	}
-	else
-# endif
-	{
+# ifdef	_NODOSDRIVE
+	strcpy(path, fullpath);
+	tr_cur[0].name = strdup2(_SS_);
+# else
+	if (dospath("", path)) tr_cur[0].name = strndup2(path, 3);
+	else {
 		strcpy(path, fullpath);
 		tr_cur[0].name = strdup2(_SS_);
 	}
-	if (Xstat(tr_cur[0].name, &st) < 0) tr_cur[0].dev = tr_cur[0].ino = 0;
+# endif
+
+	if (Xstat(nodospath(tmp, tr_cur[0].name), &st) < 0)
+		tr_cur[0].dev = tr_cur[0].ino = 0;
 	else {
 		tr_cur[0].dev = st.st_dev;
 		tr_cur[0].ino = st.st_ino;
@@ -727,11 +734,11 @@ static char *NEAR _tree(VOID_A)
 	cwd = path + 2;
 #else	/* !MSDOS */
 	cwd = path;
-#ifndef	_NODOSDRIVE
+# ifndef	_NODOSDRIVE
 	if (_dospath(cwd)) cwd += 2;
-#endif
+# endif
 #endif	/* !MSDOS */
-	if (!strcmp(cwd, _SS_));
+	if (!strcmp(cwd, _SS_)) /*EMPTY*/;
 	else for (cp = cwd; (cp = strdelim(cp, 0)); cp++, tr_line++)
 		if ((tr_line + 1) * DIRFIELD + 2 > TREEFIELD
 		|| !(tr_cur = &(tr_cur -> sub[0]))) break;
@@ -741,7 +748,7 @@ static char *NEAR _tree(VOID_A)
 		tr_line = LFILEBOTTOM - 2;
 	}
 
-	tr_scr = malloc2((FILEPERLOW - 1) * (TREEFIELD + 1));
+	tr_scr = malloc2((FILEPERROW - 1) * (TREEFIELD + 1));
 	searchtree();
 	showtree();
 	win_x = 0;
@@ -786,6 +793,10 @@ int cleanup, *ddp;
 {
 	char *path, *dupfullpath;
 
+	if (FILEPERROW < WFILEMINTREE) {
+		warning(0, NOROW_K);
+		return(NULL);
+	}
 	dupfullpath = strdup2(fullpath);
 	do {
 		path = _tree();
